#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
from hal import frccan
import struct
import warnings
import weakref

from .livewindowsendable import LiveWindowSendable
from .motorsafety import MotorSafety
from .resource import Resource
from .timer import Timer

from . import _canjaguar as _cj

__all__ = ["CANJaguar"]

def _packPercentage(value):
    if value < -1.0:
        value = -1.0
    if value > 1.0:
        value = 1.0
    return [x for x in struct.pack("<h", int(value * 32767.0))]

def _packFXP8_8(value):
    return [x for x in struct.pack("<h", int(value * 256.0))]

def _packFXP16_16(value):
    return [x for x in struct.pack("<i", int(value * 65536.0))]

def _packINT16(value):
    return [x for x in struct.pack("<h", int(value))]

def _packINT32(value):
    return [x for x in struct.pack("<i", int(value))]

def _unpackPercentage(buffer):
    return struct.unpack("<h", bytes(buffer[:2]))[0] / 32767.0

def _unpackFXP8_8(buffer):
    return struct.unpack("<h", bytes(buffer[:2]))[0] / 256.0

def _unpackFXP16_16(buffer):
    return struct.unpack("<i", bytes(buffer[:4]))[0] / 65536.0

def _unpackINT16(buffer):
    return struct.unpack("<h", bytes(buffer[:2]))[0]

def _unpackINT32(buffer):
    return struct.unpack("<i", bytes(buffer[:4]))[0]

def _FXP8_EQ(a, b):
    """Compare floats for equality as fixed point numbers"""
    return int(a * 256.0) == int(b * 256.0)

def _FXP16_EQ(a, b):
    """Compare floats for equality as fixed point numbers"""
    return int(a * 65536.0) == int(b * 65536.0)

def _sendMessageHelper(messageID, data, period):
    if (CANJaguar.kFullMessageIDMask & messageID) in CANJaguar.kTrustedMessages:
        # Make sure the data will still fit after adjusting for the token.
        if data is not None and len(data) > CANJaguar.kMaxMessageDataSize - 2:
            raise RuntimeError("CAN message has too much data.")

        trustedData = [0, 0] # token placeholder
        if data is not None:
            trustedData.extend(data)
        frccan.CANSessionMux_sendMessage(messageID, trustedData, period)
    else:
        frccan.CANSessionMux_sendMessage(messageID, data, period)

def _freeJaguar(deviceNumber, controlMode):
    # Cancel periodic messages to the Jaguar, effectively disabling it.
    # Disable periodic setpoints

    CANJaguar.allocated.free(deviceNumber-1)

    if controlMode == CANJaguar.ControlMode.PercentVbus:
        messageID = deviceNumber | _cj.LM_API_VOLT_T_SET
    elif controlMode == CANJaguar.ControlMode.Speed:
        messageID = deviceNumber | _cj.LM_API_SPD_T_SET
    elif controlMode == CANJaguar.ControlMode.Position:
        messageID = deviceNumber | _cj.LM_API_POS_T_SET
    elif controlMode == CANJaguar.ControlMode.Current:
        messageID = deviceNumber | _cj.LM_API_ICTRL_T_SET
    elif controlMode == CANJaguar.ControlMode.Voltage:
        messageID = deviceNumber | _cj.LM_API_VCOMP_T_SET
    else:
        return

    frccan.CANSessionMux_sendMessage(messageID, None,
                                     frccan.CAN_SEND_PERIOD_STOP_REPEATING)

    data = _packFXP8_8(CANJaguar.kApproxBusVoltage)
    _sendMessageHelper(_cj.LM_API_CFG_MAX_VOUT | deviceNumber, data, frccan.CAN_SEND_PERIOD_NO_REPEAT)

class CANJaguar(LiveWindowSendable, MotorSafety):
    """Texas Instruments Jaguar Speed Controller as a CAN device."""

    kMaxMessageDataSize = 8

    # The internal PID control loop in the Jaguar runs at 1kHz.
    kControllerRate = 1000
    kApproxBusVoltage = 12.0

    kReceiveStatusAttempts = 50

    allocated = Resource(63)

    kFullMessageIDMask = \
            _cj.CAN_MSGID_API_M | _cj.CAN_MSGID_MFR_M | _cj.CAN_MSGID_DTYPE_M
    kSendMessagePeriod = 20

    kTrustedMessages = set([
            _cj.LM_API_VOLT_T_EN, _cj.LM_API_VOLT_T_SET, _cj.LM_API_SPD_T_EN,
            _cj.LM_API_SPD_T_SET, _cj.LM_API_VCOMP_T_EN, _cj.LM_API_VCOMP_T_SET,
            _cj.LM_API_POS_T_EN, _cj.LM_API_POS_T_SET, _cj.LM_API_ICTRL_T_EN,
            _cj.LM_API_ICTRL_T_SET])

    class Mode:
        """Control Mode."""
        
        #: Sets an encoder as the speed reference only.
        kEncoder = 0
        
        #: Sets a quadrature encoder as the position and speed reference.
        kQuadEncoder = 1
        
        #: Sets a potentiometer as the position reference only.
        kPotentiometer = 2

    class ControlMode:
        """Determines how the Jaguar is controlled, used internally."""
        PercentVbus = 0
        Current = 1
        Speed = 2
        Position = 3
        Voltage = 4

    kCurrentFault = 1
    kTemperatureFault = 2
    kBusVoltageFault = 4
    kGateDriverFault = 8

    # Limit switch masks
    kForwardLimit = 1
    kReverseLimit = 2

    class NeutralMode:
        """Determines how the Jaguar behaves when sending a zero signal."""
        
        #: Use the NeutralMode that is set by the jumper wire on the CAN device
        Jumper = 0
        
        #: Stop the motor's rotation by applying a force.
        Brake = 1
        
        #: Do not attempt to stop the motor. Instead allow it to coast
        #: to a stop without applying resistance.
        Coast = 2

    class LimitMode:
        """Determines which sensor to use for position reference.
        Limit switches will always be used to limit the rotation. This can
        not be disabled.
        """
        
        #: Disables the soft position limits and only uses
        #: the limit switches to limit rotation.  See `getForwardLimitOK`
        #: and `getReverseLimitOK`.
        SwitchInputsOnly = 0
        
        #: Enables the soft position limits on the Jaguar.
        #: These will be used in addition to the limit switches. This does
        #: not disable the behavior of the limit switch input.
        #: See `configSoftPositionLimits`.
        SoftPositionLimits = 1

    def __init__(self, deviceNumber):
        """Constructor for the CANJaguar device.
        
        By default the device is configured in Percent mode.
        The control mode can be changed by calling one of the control modes.

        :param deviceNumber: The address of the Jaguar on the CAN bus.
        """
        MotorSafety.__init__(self)

        try:
            CANJaguar.allocated.allocate(self, deviceNumber-1)
        except IndexError as e:
            raise IndexError("CANJaguar device %d in use (increment index by one)" % deviceNumber) from e

        self.deviceNumber = deviceNumber
        self.value = 0.0

        # Parameters/configuration
        self.controlMode = CANJaguar.ControlMode.PercentVbus
        self.speedReference = _cj.LM_REF_NONE
        self.positionReference = _cj.LM_REF_NONE
        self.p = 0.0
        self.i = 0.0
        self.d = 0.0
        self.neutralMode = CANJaguar.NeutralMode.Jumper
        self.encoderCodesPerRev = 0
        self.potentiometerTurns = 0
        self.limitMode = CANJaguar.LimitMode.SwitchInputsOnly
        self.forwardLimit = 0.0
        self.reverseLimit = 0.0
        self.maxOutputVoltage = CANJaguar.kApproxBusVoltage
        self.voltageRampRate = 0.0
        self.faultTime = 0.0

        # Which parameters have been verified since they were last set?
        self.controlModeVerified = True
        self.speedRefVerified = True
        self.posRefVerified = True
        self.pVerified = True
        self.iVerified = True
        self.dVerified = True
        self.neutralModeVerified = True
        self.encoderCodesPerRevVerified = True
        self.potentiometerTurnsVerified = True
        self.forwardLimitVerified = True
        self.reverseLimitVerified = True
        self.limitModeVerified = True
        self.maxOutputVoltageVerified = True
        self.voltageRampRateVerified = True
        self.faultTimeVerified = True

        # Status data
        self.busVoltage = 0.0
        self.outputVoltage = 0.0
        self.outputCurrent = 0.0
        self.temperature = 0.0
        self.position = 0.0
        self.speed = 0.0
        self.limits = 0
        self.faults = 0
        self.firmwareVersion = 0
        self.hardwareVersion = 0

        # Which periodic status messages have we received at least once?
        self.receivedStatusMessage0 = False
        self.receivedStatusMessage1 = False
        self.receivedStatusMessage2 = False

        self.controlEnabled = True

        receivedFirmwareVersion = False

        # Request firmware and hardware version only once
        self.requestMessage(frccan.CAN_IS_FRAME_REMOTE |
                            _cj.CAN_MSGID_API_FIRMVER)
        self.requestMessage(_cj.LM_API_HWVER)

        # Establish finalizer
        self._canjaguar_finalizer = weakref.finalize(self, _freeJaguar,
                                                     self.deviceNumber,
                                                     self.controlMode)
        
        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

        # Wait until we've gotten all of the status data at least once.
        for i in range(CANJaguar.kReceiveStatusAttempts):
            Timer.delay(0.001)

            self.setupPeriodicStatus()
            self.updatePeriodicStatus()

            if not receivedFirmwareVersion:
                try:
                    data = self.getMessage(_cj.CAN_MSGID_API_FIRMVER,
                                           _cj.CAN_MSGID_FULL_M)
                    self.firmwareVersion = _unpackINT32(data)
                    receivedFirmwareVersion = True
                except frccan.CANMessageNotFound:
                    pass

            if (self.receivedStatusMessage0 and
                self.receivedStatusMessage1 and
                self.receivedStatusMessage2 and
                receivedFirmwareVersion):
                break
        else:
            raise frccan.CANMessageNotFound("message not found")

        try:
            data = self.getMessage(_cj.LM_API_HWVER, _cj.CAN_MSGID_FULL_M)
            self.hardwareVersion = data[0]
        except frccan.CANError:
            # Not all Jaguar firmware reports a hardware version.
            self.hardwareVersion = 0

        # 3330 was the first shipping RDK firmware version for the Jaguar
        if self.firmwareVersion >= 3330 or self.firmwareVersion < 108:
            from .driverstation import DriverStation
            if self.firmwareVersion < 3330:
                DriverStation.reportError("Jag %d firmware %d is too old (must be at least version 108 of the FIRST approved firmware)" % (self.deviceNumber, self.firmwareVersion), False)
            else:
                DriverStation.reportError("Jag %d firmware %d is not FIRST approved (must be at least version 108 of the FIRST approved firmware)" % (self.deviceNumber, self.firmwareVersion), False)

    def free(self):
        """
        Cancel periodic messages to the Jaguar, effectively disabling it.
        No other methods should be called after this is called.
        """
        self._canjaguar_finalizer()

    def getDeviceNumber(self):
        """:returns: The CAN ID passed in the constructor
        """
        return self.deviceNumber

    def get(self):
        """Get the recently set outputValue set point.

        The scale and the units depend on the mode the Jaguar is in.
        
        - In percentVbus mode, the outputValue is from -1.0 to 1.0 (same as
          PWM Jaguar).
        - In voltage mode, the outputValue is in volts.
        - In current mode, the outputValue is in amps.
        - In speed mode, the outputValue is in rotations/minute.
        - In position mode, the outputValue is in rotations.

        :returns: The most recently set outputValue set point.
        """
        return self.value

    def set(self, outputValue, syncGroup=0):
        """Sets the output set-point value.

        The scale and the units depend on the mode the Jaguar is in.
        
        - In percentVbus Mode, the outputValue is from -1.0 to 1.0 (same as
          PWM Jaguar).
        - In voltage Mode, the outputValue is in volts.
        - In current Mode, the outputValue is in amps.
        - In speed mode, the outputValue is in rotations/minute.
        - In position Mode, the outputValue is in rotations.

        :param outputValue: The set-point to sent to the motor controller.
        :param syncGroup: The update group to add this set() to, pending
            UpdateSyncGroup().  If 0 (default), update immediately.
        """
        if self.controlEnabled:
            if self.controlMode == self.ControlMode.PercentVbus:
                messageID = _cj.LM_API_VOLT_T_SET
                data = _packPercentage(outputValue)
            elif self.controlMode == self.ControlMode.Speed:
                messageID = _cj.LM_API_SPD_T_SET
                data = _packFXP16_16(outputValue)
            elif self.controlMode == self.ControlMode.Position:
                messageID = _cj.LM_API_POS_T_SET
                data = _packFXP16_16(outputValue)
            elif self.controlMode == self.ControlMode.Current:
                messageID = _cj.LM_API_ICTRL_T_SET
                data = _packFXP8_8(outputValue)
            elif self.controlMode == self.ControlMode.Voltage:
                messageID = _cj.LM_API_VCOMP_T_SET
                data = _packFXP8_8(outputValue)
            else:
                return

            if syncGroup != 0:
                data.append(syncGroup)

            self.sendMessage(messageID, data, self.kSendMessagePeriod)

            self.feed()

        self.value = outputValue

        self.verify()

    def verify(self):
        """Check all unverified params and make sure they're equal to their
        local cached versions. If a value isn't available, it gets requested.
        If a value doesn't match up, it gets set again.
        """
        # If the Jaguar lost power, everything should be considered unverified
        try:
            data = self.getMessage(_cj.LM_API_STATUS_POWER,
                                   _cj.CAN_MSGID_FULL_M)
            if data[0] != 0: # power cycled
                # Clear the power cycled bit
                data[0] = 1
                self.sendMessage(_cj.LM_API_STATUS_POWER, data[:1])

                # Mark everything as unverified
                self.controlModeVerified = False
                self.speedRefVerified = False
                self.posRefVerified = False
                self.neutralModeVerified = False
                self.encoderCodesPerRevVerified = False
                self.potentiometerTurnsVerified = False
                self.forwardLimitVerified = False
                self.reverseLimitVerified = False
                self.limitModeVerified = False
                self.maxOutputVoltageVerified = False
                self.faultTimeVerified = False

                if (self.controlMode == self.ControlMode.PercentVbus or
                    self.controlMode == self.ControlMode.Voltage):
                    self.voltageRampRateVerified = False
                else:
                    self.pVerified = False
                    self.iVerified = False
                    self.dVerified = False

                # Verify periodic status messages again
                self.receivedStatusMessage0 = False
                self.receivedStatusMessage1 = False
                self.receivedStatusMessage2 = False

                # Remove any old values from netcomms. Otherwise, parameters
                # are incorrectly marked as verified based on stale messages.
                messages = [
                    _cj.LM_API_SPD_REF, _cj.LM_API_POS_REF,
                    _cj.LM_API_SPD_PC, _cj.LM_API_POS_PC,
                    _cj.LM_API_ICTRL_PC, _cj.LM_API_SPD_IC,
                    _cj.LM_API_POS_IC, _cj.LM_API_ICTRL_IC,
                    _cj.LM_API_SPD_DC, _cj.LM_API_POS_DC,
                    _cj.LM_API_ICTRL_DC, _cj.LM_API_CFG_ENC_LINES,
                    _cj.LM_API_CFG_POT_TURNS, _cj.LM_API_CFG_BRAKE_COAST,
                    _cj.LM_API_CFG_LIMIT_MODE, _cj.LM_API_CFG_LIMIT_REV,
                    _cj.LM_API_CFG_MAX_VOUT, _cj.LM_API_VOLT_SET_RAMP,
                    _cj.LM_API_VCOMP_COMP_RAMP, _cj.LM_API_CFG_FAULT_TIME,
                    _cj.LM_API_CFG_LIMIT_FWD]

                for message in messages:
                    try:
                        data = self.getMessage(message, _cj.CAN_MSGID_FULL_M)
                    except frccan.CANMessageNotFound:
                        pass
        except frccan.CANMessageNotFound:
            self.requestMessage(_cj.LM_API_STATUS_POWER)

        # Verify that any recently set parameters are correct
        if not self.controlModeVerified and self.controlEnabled:
            try:
                data = self.getMessage(_cj.LM_API_STATUS_CMODE,
                                       _cj.CAN_MSGID_FULL_M)
                mode = data[0]
                if self.controlMode == mode:
                    self.controlModeVerified = True
                else:
                    # Enable control again to resend the control mode
                    self.enableControl()
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_STATUS_CMODE)

        if not self.speedRefVerified:
            try:
                data = self.getMessage(_cj.LM_API_SPD_REF, _cj.CAN_MSGID_FULL_M)
                speedRef = data[0]
                if self.speedReference == speedRef:
                    self.speedRefVerified = True
                else:
                    # It's wrong - set it again
                    self.setSpeedReference(self.speedReference)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_SPD_REF)

        if not self.posRefVerified:
            try:
                data = self.getMessage(_cj.LM_API_POS_REF, _cj.CAN_MSGID_FULL_M)
                posRef = data[0]
                if self.positionReference == posRef:
                    self.posRefVerified = True
                else:
                    # It's wrong - set it again
                    self.setPositionReference(self.positionReference)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_POS_REF)

        if not self.pVerified:
            message = 0
            if self.controlMode == self.ControlMode.Speed:
                message = _cj.LM_API_SPD_PC
            elif self.controlMode == self.ControlMode.Position:
                message = _cj.LM_API_POS_PC
            elif self.controlMode == self.ControlMode.Current:
                message = _cj.LM_API_ICTRL_PC

            if message != 0:
                try:
                    data = self.getMessage(message, _cj.CAN_MSGID_FULL_M)
                    p = _unpackFXP16_16(data)
                    if _FXP16_EQ(self.p, p):
                        self.pVerified = True
                    else:
                        # It's wrong - set it again
                        self.setP(self.p)
                except frccan.CANMessageNotFound:
                    # Verification is needed but not available - request it again.
                    self.requestMessage(message)

        if not self.iVerified:
            message = 0
            if self.controlMode == self.ControlMode.Speed:
                message = _cj.LM_API_SPD_IC
            elif self.controlMode == self.ControlMode.Position:
                message = _cj.LM_API_POS_IC
            elif self.controlMode == self.ControlMode.Current:
                message = _cj.LM_API_ICTRL_IC

            if message != 0:
                try:
                    data = self.getMessage(message, _cj.CAN_MSGID_FULL_M)
                    i = _unpackFXP16_16(data)
                    if _FXP16_EQ(self.i, i):
                        self.iVerified = True
                    else:
                        # It's wrong - set it again
                        self.setI(self.i)
                except frccan.CANMessageNotFound:
                    # Verification is needed but not available - request it again.
                    self.requestMessage(message)

        if not self.dVerified:
            message = 0
            if self.controlMode == self.ControlMode.Speed:
                message = _cj.LM_API_SPD_DC
            elif self.controlMode == self.ControlMode.Position:
                message = _cj.LM_API_POS_DC
            elif self.controlMode == self.ControlMode.Current:
                message = _cj.LM_API_ICTRL_DC

            if message != 0:
                try:
                    data = self.getMessage(message, _cj.CAN_MSGID_FULL_M)
                    d = _unpackFXP16_16(data)
                    if _FXP16_EQ(self.d, d):
                        self.dVerified = True
                    else:
                        # It's wrong - set it again
                        self.setD(self.d)
                except frccan.CANMessageNotFound:
                    # Verification is needed but not available - request it again.
                    self.requestMessage(message)

        if not self.neutralModeVerified:
            try:
                data = self.getMessage(_cj.LM_API_CFG_BRAKE_COAST,
                                       _cj.CAN_MSGID_FULL_M)
                mode = data[0]
                if mode == self.neutralMode:
                    self.neutralModeVerified = True
                else:
                    #It's wrong - set it again
                    self.configNeutralMode(self.neutralMode)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_CFG_BRAKE_COAST)

        if not self.encoderCodesPerRevVerified:
            try:
                data = self.getMessage(_cj.LM_API_CFG_ENC_LINES,
                                _cj.CAN_MSGID_FULL_M)
                codes = _unpackINT16(data)
                if codes == self.encoderCodesPerRev:
                    self.encoderCodesPerRevVerified = True
                else:
                    #It's wrong - set it again
                    self.configEncoderCodesPerRev(self.encoderCodesPerRev)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_CFG_ENC_LINES)

        if not self.potentiometerTurnsVerified:
            try:
                data = self.getMessage(_cj.LM_API_CFG_POT_TURNS,
                                       _cj.CAN_MSGID_FULL_M)
                turns = _unpackINT16(data)
                if turns == self.potentiometerTurns:
                    self.potentiometerTurnsVerified = True
                else:
                    #It's wrong - set it again
                    self.configPotentiometerTurns(self.potentiometerTurns)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_CFG_POT_TURNS)

        if not self.limitModeVerified:
            try:
                data = self.getMessage(_cj.LM_API_CFG_LIMIT_MODE,
                                       _cj.CAN_MSGID_FULL_M)
                mode = data[0]
                if mode == self.limitMode:
                    self.limitModeVerified = True
                else:
                    #It's wrong - set it again
                    self.configLimitMode(self.limitMode)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_CFG_LIMIT_MODE)

        if not self.forwardLimitVerified:
            try:
                data = self.getMessage(_cj.LM_API_CFG_LIMIT_FWD,
                                       _cj.CAN_MSGID_FULL_M)
                limit = _unpackFXP16_16(data)
                if _FXP16_EQ(limit, self.forwardLimit):
                    self.forwardLimitVerified = True
                else:
                    #It's wrong - set it again
                    self.configForwardLimit(self.forwardLimit)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_CFG_LIMIT_FWD)

        if not self.reverseLimitVerified:
            try:
                data = self.getMessage(_cj.LM_API_CFG_LIMIT_REV,
                                       _cj.CAN_MSGID_FULL_M)
                limit = _unpackFXP16_16(data)
                if _FXP16_EQ(limit, self.reverseLimit):
                    self.reverseLimitVerified = True
                else:
                    #It's wrong - set it again
                    self.configReverseLimit(self.reverseLimit)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_CFG_LIMIT_REV)

        if not self.maxOutputVoltageVerified:
            try:
                data = self.getMessage(_cj.LM_API_CFG_MAX_VOUT,
                                       _cj.CAN_MSGID_FULL_M)
                voltage = _unpackFXP8_8(data)
                # The returned max output voltage is sometimes slightly higher
                # or lower than what was sent.  This should not trigger
                # resending the message.
                if abs(voltage - self.maxOutputVoltage) < 0.1:
                    self.maxOutputVoltageVerified = True
                else:
                    # It's wrong - set it again
                    self.configMaxOutputVoltage(self.maxOutputVoltage)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_CFG_MAX_VOUT)

        if not self.voltageRampRateVerified:
            if self.controlMode == CANJaguar.ControlMode.PercentVbus:
                try:
                    data = self.getMessage(_cj.LM_API_VOLT_SET_RAMP,
                                           _cj.CAN_MSGID_FULL_M)
                    rate = _unpackPercentage(data)
                    if _FXP16_EQ(rate, self.voltageRampRate):
                        self.voltageRampRateVerified = True
                    else:
                        # It's wrong - set it again
                        self.setVoltageRampRate(self.voltageRampRate)
                except frccan.CANMessageNotFound:
                    # Verification is needed but not available - request it again.
                    self.requestMessage(_cj.LM_API_VOLT_SET_RAMP)
            elif self.controlMode == CANJaguar.ControlMode.Voltage:
                try:
                    data = self.getMessage(_cj.LM_API_VCOMP_COMP_RAMP,
                                           _cj.CAN_MSGID_FULL_M)
                    rate = _unpackFXP8_8(data)
                    if _FXP8_EQ(rate, self.voltageRampRate):
                        self.voltageRampRateVerified = True
                    else:
                        # It's wrong - set it again
                        self.setVoltageRampRate(self.voltageRampRate)
                except frccan.CANMessageNotFound:
                    # Verification is needed but not available - request it again.
                    self.requestMessage(_cj.LM_API_VCOMP_COMP_RAMP)

        if not self.faultTimeVerified:
            try:
                data = self.getMessage(_cj.LM_API_CFG_FAULT_TIME,
                                       _cj.CAN_MSGID_FULL_M)
                faultTime = _unpackINT16(data)
                if int(self.faultTime * 1000.0) == faultTime:
                    self.faultTimeVerified = True
                else:
                    #It's wrong - set it again
                    self.configFaultTime(self.faultTime)
            except frccan.CANMessageNotFound:
                # Verification is needed but not available - request it again.
                self.requestMessage(_cj.LM_API_CFG_FAULT_TIME)

        if (not self.receivedStatusMessage0 or
            not self.receivedStatusMessage1 or
            not self.receivedStatusMessage2):
            # If the periodic status messages haven't been verified as
            # received, request periodic status messages again and attempt
            # to unpack any available ones.
            self.setupPeriodicStatus()
            self.getTemperature()
            self.getPosition()
            self.getFaults()

    def disable(self):
        """Common interface for disabling a motor.

        .. deprecated :: 2015
            Use :func:`disableControl` instead.
        """
        warnings.warn("use disableControl instead", DeprecationWarning)
        self.disableControl()

    # PIDOutput interface
    def pidWrite(self, output):
        if self.controlMode == self.ControlMode.PercentVbus:
            self.set(output)
        else:
            raise ValueError("PID only supported in PercentVbus mode")

    def setSpeedReference(self, reference):
        """Set the reference source device for speed controller mode.

        Choose encoder as the source of speed feedback when in speed control
        mode.

        :param reference: Specify a speed reference.
        """
        self.sendMessage(_cj.LM_API_SPD_REF, [reference])
        self.speedReference = reference
        self.speedRefVerified = False

    def setPositionReference(self, reference):
        """Set the reference source device for position controller mode.

        Choose between using and encoder and using a potentiometer
        as the source of position feedback when in position control mode.

        :param reference: Specify a position reference.
        """
        self.sendMessage(_cj.LM_API_POS_REF, [reference])
        self.positionReference = reference
        self.posRefVerified = False

    def setP(self, p):
        """Set the P constant for the closed loop modes.

        :param p: The proportional gain of the Jaguar's PID controller.
        """
        data = _packFXP16_16(p)

        if self.controlMode == self.ControlMode.Speed:
            self.sendMessage(_cj.LM_API_SPD_PC, data)
        elif self.controlMode == self.ControlMode.Position:
            self.sendMessage(_cj.LM_API_POS_PC, data)
        elif self.controlMode == self.ControlMode.Current:
            self.sendMessage(_cj.LM_API_ICTRL_PC, data)
        else:
            raise ValueError("PID constants only apply in Speed, Position, and Current mode")

        self.p = p
        self.pVerified = False

    def setI(self, i):
        """Set the I constant for the closed loop modes.

        :param i: The integral gain of the Jaguar's PID controller.
        """
        data = _packFXP16_16(i)

        if self.controlMode == self.ControlMode.Speed:
            self.sendMessage(_cj.LM_API_SPD_IC, data)
        elif self.controlMode == self.ControlMode.Position:
            self.sendMessage(_cj.LM_API_POS_IC, data)
        elif self.controlMode == self.ControlMode.Current:
            self.sendMessage(_cj.LM_API_ICTRL_IC, data)
        else:
            raise ValueError("PID constants only apply in Speed, Position, and Current mode")

        self.i = i
        self.iVerified = False

    def setD(self, d):
        """Set the D constant for the closed loop modes.

        :param d: The derivative gain of the Jaguar's PID controller.
        """
        data = _packFXP16_16(d)

        if self.controlMode == self.ControlMode.Speed:
            self.sendMessage(_cj.LM_API_SPD_DC, data)
        elif self.controlMode == self.ControlMode.Position:
            self.sendMessage(_cj.LM_API_POS_DC, data)
        elif self.controlMode == self.ControlMode.Current:
            self.sendMessage(_cj.LM_API_ICTRL_DC, data)
        else:
            raise ValueError("PID constants only apply in Speed, Position, and Current mode")

        self.d = d
        self.dVerified = False

    def setPID(self, p, i, d):
        """Set the P, I, and D constants for the closed loop modes.

        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.setP(p)
        self.setI(i)
        self.setD(d)

    def getP(self):
        """Get the Proportional gain of the controller.

        :returns: The proportional gain.
        """
        if (self.controlMode == self.ControlMode.PercentVbus or
            self.controlMode == self.ControlMode.Voltage):
            raise ValueError("PID does not apply in Percent or Voltage control modes")
        return self.p

    def getI(self):
        """Get the Integral gain of the controller.

        :returns: The integral gain.
        """
        if (self.controlMode == self.ControlMode.PercentVbus or
            self.controlMode == self.ControlMode.Voltage):
            raise ValueError("PID does not apply in Percent or Voltage control modes")
        return self.i

    def getD(self):
        """Get the Derivative gain of the controller.

        :returns: The derivative gain.
        """
        if (self.controlMode == self.ControlMode.PercentVbus or
            self.controlMode == self.ControlMode.Voltage):
            raise ValueError("PID does not apply in Percent or Voltage control modes")
        return self.d

    def enableControl(self, encoderInitialPosition=0.0):
        """Enable the closed loop controller.

        Start actually controlling the output based on the feedback.
        If starting a position controller with an encoder reference,
        use the encoderInitialPosition parameter to initialize the
        encoder state.

        :param encoderInitialPosition: Encoder position to set if position
            with encoder reference (default of 0.0).  Ignored otherwise.
        """
        if self.controlMode == self.ControlMode.PercentVbus:
            self.sendMessage(_cj.LM_API_VOLT_T_EN, None)
        elif self.controlMode == self.ControlMode.Speed:
            self.sendMessage(_cj.LM_API_SPD_T_EN, None)
        elif self.controlMode == self.ControlMode.Position:
            data = _packFXP16_16(encoderInitialPosition)
            self.sendMessage(_cj.LM_API_POS_T_EN, data)
        elif self.controlMode == self.ControlMode.Current:
            self.sendMessage(_cj.LM_API_ICTRL_T_EN, None)
        elif self.controlMode == self.ControlMode.Voltage:
            self.sendMessage(_cj.LM_API_VCOMP_T_EN, None)

        self.controlEnabled = True

    def disableControl(self):
        """Disable the closed loop controller.

        Stop driving the output based on the feedback.
        """
        # Disable all control modes.
        self.sendMessage(_cj.LM_API_VOLT_DIS, None)
        self.sendMessage(_cj.LM_API_SPD_DIS, None)
        self.sendMessage(_cj.LM_API_POS_DIS, None)
        self.sendMessage(_cj.LM_API_ICTRL_DIS, None)
        self.sendMessage(_cj.LM_API_VCOMP_DIS, None)

        # Stop all periodic setpoints
        self.sendMessage(_cj.LM_API_VOLT_T_SET, None,
                         frccan.CAN_SEND_PERIOD_STOP_REPEATING)
        self.sendMessage(_cj.LM_API_SPD_T_SET, None,
                         frccan.CAN_SEND_PERIOD_STOP_REPEATING)
        self.sendMessage(_cj.LM_API_POS_T_SET, None,
                         frccan.CAN_SEND_PERIOD_STOP_REPEATING)
        self.sendMessage(_cj.LM_API_ICTRL_T_SET, None,
                         frccan.CAN_SEND_PERIOD_STOP_REPEATING)
        self.sendMessage(_cj.LM_API_VCOMP_T_SET, None,
                         frccan.CAN_SEND_PERIOD_STOP_REPEATING)

        self.controlEnabled = False

    def setPercentMode(self):
        """Enable controlling the motor voltage as a percentage of the bus
        voltage without any position or speed feedback.

        After calling this you must call :func:`enableControl` to enable
        the device.
        """
        self.changeControlMode(self.ControlMode.PercentVbus)
        self.setPositionReference(_cj.LM_REF_NONE)
        self.setSpeedReference(_cj.LM_REF_NONE)

    def setPercentModeEncoder(self, codesPerRev):
        """Enable controlling the motor voltage as a percentage of the bus
        voltage, and enable speed sensing from a non-quadrature encoder.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param codesPerRev: The counts per revolution on the encoder
        """
        self.changeControlMode(self.ControlMode.PercentVbus)
        self.setPositionReference(_cj.LM_REF_NONE)
        self.setSpeedReference(_cj.LM_REF_ENCODER)
        self.configEncoderCodesPerRev(codesPerRev)

    def setPercentModeQuadEncoder(self, codesPerRev):
        """Enable controlling the motor voltage as a percentage of the bus
        voltage, and enable position and speed sensing from a quadrature
        encoder.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param tag: The constant {@link CANJaguar#kQuadEncoder}
        :param codesPerRev: The counts per revolution on the encoder
        """
        self.changeControlMode(self.ControlMode.PercentVbus)
        self.setPositionReference(_cj.LM_REF_ENCODER)
        self.setSpeedReference(_cj.LM_REF_QUAD_ENCODER)
        self.configEncoderCodesPerRev(codesPerRev)

    def setPercentModePotentiometer(self):
        """Enable controlling the motor voltage as a percentage of the bus
        voltage, and enable position sensing from a potentiometer and no
        speed feedback.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param tag: The constant {@link CANJaguar#kPotentiometer}
        """
        self.changeControlMode(self.ControlMode.PercentVbus)
        self.setPositionReference(_cj.LM_REF_POT)
        self.setSpeedReference(_cj.LM_REF_NONE)
        self.configPotentiometerTurns(1)

    def setCurrentModePID(self, p, i, d):
        """Enable controlling the motor current with a PID loop.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.changeControlMode(self.ControlMode.Current)
        self.setPositionReference(_cj.LM_REF_NONE)
        self.setSpeedReference(_cj.LM_REF_NONE)
        self.setPID(p, i, d)

    def setCurrentModeEncoder(self, codesPerRev, p, i, d):
        """Enable controlling the motor current with a PID loop, and enable
        speed sensing from a non-quadrature encoder.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.changeControlMode(self.ControlMode.Current)
        self.setPositionReference(_cj.LM_REF_NONE)
        self.setSpeedReference(_cj.LM_REF_NONE)
        self.configEncoderCodesPerRev(codesPerRev)
        self.setPID(p, i, d)

    def setCurrentModeQuadEncoder(self, codesPerRev, p, i, d):
        """Enable controlling the motor current with a PID loop, and enable
        speed and position sensing from a quadrature encoder.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param codesPerRev: The counts per revolution on the encoder
        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.changeControlMode(self.ControlMode.Current)
        self.setPositionReference(_cj.LM_REF_ENCODER)
        self.setSpeedReference(_cj.LM_REF_QUAD_ENCODER)
        self.configEncoderCodesPerRev(codesPerRev)
        self.setPID(p, i, d)

    def setCurrentModePotentiometer(self, p, i, d):
        """Enable controlling the motor current with a PID loop, and enable
        position sensing from a potentiometer.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.changeControlMode(self.ControlMode.Current)
        self.setPositionReference(_cj.LM_REF_POT)
        self.setSpeedReference(_cj.LM_REF_NONE)
        self.configPotentiometerTurns(1)
        self.setPID(p, i, d)

    def setSpeedModeEncoder(self, codesPerRev, p, i, d):
        """Enable controlling the speed with a feedback loop from a
        non-quadrature encoder.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param codesPerRev: The counts per revolution on the encoder
        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.changeControlMode(self.ControlMode.Speed)
        self.setPositionReference(_cj.LM_REF_NONE)
        self.setSpeedReference(_cj.LM_REF_ENCODER)
        self.configEncoderCodesPerRev(codesPerRev)
        self.setPID(p, i, d)

    def setSpeedModeQuadEncoder(self, codesPerRev, p, i, d):
        """Enable controlling the speed with a feedback loop from a
        quadrature encoder.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param codesPerRev: The counts per revolution on the encoder
        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.changeControlMode(self.ControlMode.Speed)
        self.setPositionReference(_cj.LM_REF_ENCODER)
        self.setSpeedReference(_cj.LM_REF_QUAD_ENCODER)
        self.configEncoderCodesPerRev(codesPerRev)
        self.setPID(p, i, d)

    def setPositionModeQuadEncoder(self, codesPerRev, p, i, d):
        """Enable controlling the position with a feedback loop using an
        encoder.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param codesPerRev: The counts per revolution on the encoder
        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.changeControlMode(self.ControlMode.Position)
        self.setPositionReference(_cj.LM_REF_ENCODER)
        self.configEncoderCodesPerRev(codesPerRev)
        self.setPID(p, i, d)

    def setPositionModePotentiometer(self, p, i, d):
        """Enable controlling the position with a feedback loop using a
        potentiometer.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param p: The proportional gain of the Jaguar's PID controller.
        :param i: The integral gain of the Jaguar's PID controller.
        :param d: The differential gain of the Jaguar's PID controller.
        """
        self.changeControlMode(self.ControlMode.Position)
        self.setPositionReference(_cj.LM_REF_POT)
        self.configPotentiometerTurns(1)
        self.setPID(p, i, d)

    def setVoltageMode(self):
        """Enable controlling the motor voltage without any position or speed
        feedback.

        After calling this you must call :func:`enableControl` to enable
        the device.
        """
        self.changeControlMode(self.ControlMode.Voltage)
        self.setPositionReference(_cj.LM_REF_NONE)
        self.setSpeedReference(_cj.LM_REF_NONE)

    def setVoltageModeEncoder(self, codesPerRev):
        """Enable controlling the motor voltage with speed feedback from a
        non-quadrature encoder and no position feedback.<br>

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param codesPerRev: The counts per revolution on the encoder
        """
        self.changeControlMode(self.ControlMode.Voltage)
        self.setPositionReference(_cj.LM_REF_NONE)
        self.setSpeedReference(_cj.LM_REF_ENCODER)
        self.configEncoderCodesPerRev(codesPerRev)

    def setVoltageModeQuadEncoder(self, codesPerRev):
        """Enable controlling the motor voltage with position and speed
        feedback from a quadrature encoder.

        After calling this you must call :func:`enableControl` to enable
        the device.

        :param tag: The constant {@link CANJaguar#kQuadEncoder}
        :param codesPerRev: The counts per revolution on the encoder
        """
        self.changeControlMode(self.ControlMode.Voltage)
        self.setPositionReference(_cj.LM_REF_ENCODER)
        self.setSpeedReference(_cj.LM_REF_QUAD_ENCODER)
        self.configEncoderCodesPerRev(codesPerRev)

    def setVoltageModePotentiometer(self):
        """Enable controlling the motor voltage with position feedback from a
        potentiometer and no speed feedback.

        After calling this you must call :func:`enableControl` to enable
        the device.
        """
        self.changeControlMode(self.ControlMode.Voltage)
        self.setPositionReference(_cj.LM_REF_POT)
        self.setSpeedReference(_cj.LM_REF_NONE)
        self.configPotentiometerTurns(1)

    def changeControlMode(self, controlMode):
        """Used internally.  In order to set the control mode see the methods
        listed below.

        Change the control mode of this Jaguar object.

        After changing modes, configure any PID constants or other settings
        needed and then EnableControl() to actually change the mode on the
        Jaguar.

        :param controlMode: The new mode.
        """
        # Disable the previous mode
        self.disableControl()

        # Update the local mode
        self.controlMode = controlMode
        self.controlModeVerified = False

        # Update the finalizer
        self._canjaguar_finalizer.detach()
        self._canjaguar_finalizer = weakref.finalize(self, _freeJaguar,
                                                     self.deviceNumber,
                                                     self.controlMode)

    def getControlMode(self):
        """Get the active control mode from the Jaguar.

        Ask the Jagaur what mode it is in.

        :return ControlMode: that the Jag is in.
        """
        return self.controlMode

    def getBusVoltage(self):
        """Get the voltage at the battery input terminals of the Jaguar.

        :returns: The bus voltage in Volts.
        """
        self.updatePeriodicStatus()
        return self.busVoltage

    def getOutputVoltage(self):
        """Get the voltage being output from the motor terminals of the Jaguar.

        :returns: The output voltage in Volts.
        """
        self.updatePeriodicStatus()
        return self.outputVoltage

    def getOutputCurrent(self):
        """Get the current through the motor terminals of the Jaguar.

        :returns: The output current in Amps.
        """
        self.updatePeriodicStatus()
        return self.outputCurrent

    def getTemperature(self):
        """Get the internal temperature of the Jaguar.

        :returns: The temperature of the Jaguar in degrees Celsius.
        """
        self.updatePeriodicStatus()
        return self.temperature

    def getPosition(self):
        """Get the position of the encoder or potentiometer.

        :returns: The position of the motor in rotations based on the
            configured feedback. See :func:`configPotentiometerTurns` and
            :func:`configEncoderCodesPerRev`.
        """
        self.updatePeriodicStatus()
        return self.position

    def getSpeed(self):
        """Get the speed of the encoder.

        :returns: The speed of the motor in RPM based on the configured
            feedback.
        """
        self.updatePeriodicStatus()
        return self.speed

    def getForwardLimitOK(self):
        """Get the status of the forward limit switch.

        :returns: True if the motor is allowed to turn in the forward direction.
        """
        self.updatePeriodicStatus()
        return (self.limits & self.kForwardLimit) != 0

    def getReverseLimitOK(self):
        """Get the status of the reverse limit switch.

        :returns: True if the motor is allowed to turn in the reverse direction.
        """
        self.updatePeriodicStatus()
        return (self.limits & self.kReverseLimit) != 0

    def getFaults(self):
        """Get the status of any faults the Jaguar has detected.

        :returns: A bit-mask of faults defined by the "Faults" constants.
        
                  - `kCurrentFault`
                  - `kBusVoltageFault`
                  - `kTemperatureFault`
                  - `GateDriverFault`
        """
        self.updatePeriodicStatus()
        return self.faults

    def setVoltageRampRate(self, rampRate):
        """Set the maximum voltage change rate.

        When in PercentVbus or Voltage output mode, the rate at which the
        voltage changes can be limited to reduce current spikes.  set this
        to 0.0 to disable rate limiting.

        :param rampRate: The maximum rate of voltage change in Percent
            Voltage mode in V/s.
        """
        if self.controlMode == self.ControlMode.PercentVbus:
            data = _packPercentage(rampRate / (self.maxOutputVoltage *
                                               self.kControllerRate))
            message = _cj.LM_API_VOLT_SET_RAMP
        elif self.controlMode == self.ControlMode.Voltage:
            data = _packFXP8_8(rampRate / self.kControllerRate)
            message = _cj.LM_API_VCOMP_COMP_RAMP
        else:
            raise ValueError("Voltage ramp rate only applies in Percentage and Voltage modes")

        self.sendMessage(message, data)

    def getFirmwareVersion(self):
        """Get the version of the firmware running on the Jaguar.

        :returns: The firmware version.  0 if the device did not respond.
        """
        return self.firmwareVersion

    def getHardwareVersion(self):
        """Get the version of the Jaguar hardware.

        :returns: The hardware version. 1: Jaguar,  2: Black Jaguar
        """
        return self.hardwareVersion

    def configNeutralMode(self, mode):
        """Configure what the controller does to the H-Bridge when neutral
        (not driving the output).

        This allows you to override the jumper configuration for brake or coast.

        :param mode: Select to use the jumper setting or to override it to
            coast or brake (see `NeutralMode`).
        """
        self.sendMessage(_cj.LM_API_CFG_BRAKE_COAST, [mode])
        self.neutralMode = mode
        self.neutralModeVerified = False

    def configEncoderCodesPerRev(self, codesPerRev):
        """Configure how many codes per revolution are generated by your
        encoder.

        :param codesPerRev: The number of counts per revolution in 1X mode.
        """
        data = _packINT16(codesPerRev)
        self.sendMessage(_cj.LM_API_CFG_ENC_LINES, data)
        self.encoderCodesPerRev = codesPerRev
        self.encoderCodesPerRevVerified = False

    def configPotentiometerTurns(self, turns):
        """Configure the number of turns on the potentiometer.

        There is no special support for continuous turn potentiometers.
        Only integer numbers of turns are supported.

        :param turns: The number of turns of the potentiometer
        """
        data = _packINT16(turns)
        self.sendMessage(_cj.LM_API_CFG_POT_TURNS, data)
        self.potentiometerTurns = turns
        self.potentiometerTurnsVerified = False

    def configSoftPositionLimits(self, forwardLimitPosition,
                                 reverseLimitPosition):
        """Configure Soft Position Limits when in Position Controller mode.

        When controlling position, you can add additional limits on top of
        the limit switch inputs that are based on the position feedback.
        If the position limit is reached or the switch is opened, that
        direction will be disabled.

        :param forwardLimitPosition: The position that, if exceeded, will
            disable the forward direction.
        :param reverseLimitPosition: The position that, if exceeded, will
            disable the reverse direction.
        """
        self.configLimitMode(self.LimitMode.SoftPositionLimits)
        self.configForwardLimit(forwardLimitPosition)
        self.configReverseLimit(reverseLimitPosition)

    def disableSoftPositionLimits(self):
        """Disable Soft Position Limits if previously enabled.<br>

        Soft Position Limits are disabled by default.
        """
        self.configLimitMode(self.LimitMode.SwitchInputsOnly)

    def configLimitMode(self, mode):
        """Set the limit mode for position control mode.<br>

        Use :func:`configSoftPositionLimits` or
        :func:`disableSoftPositionLimits` to set this automatically.

        :param mode: The `LimitMode` to use to limit the rotation of the device.
        """
        self.sendMessage(_cj.LM_API_CFG_LIMIT_MODE, [mode])

    def configForwardLimit(self, forwardLimitPosition):
        """Set the position that, if exceeded, will disable the forward
        direction.

        Use :func:`configSoftPositionLimits` to set this and the
        `LimitMode` automatically.

        :param forwardLimitPosition: The position that, if exceeded, will
            disable the forward direction.
        """
        data = _packFXP16_16(forwardLimitPosition)
        data.append(1)
        self.sendMessage(_cj.LM_API_CFG_LIMIT_FWD, data)

        self.forwardLimit = forwardLimitPosition
        self.forwardLimitVerified = False

    def configReverseLimit(self, reverseLimitPosition):
        """Set the position that, if exceeded, will disable the reverse
        direction.

        Use :func:`configSoftPositionLimits` to set this and the
        `LimitMode` automatically.

        :param reverseLimitPosition: The position that, if exceeded, will
            disable the reverse direction.
        """
        data = _packFXP16_16(reverseLimitPosition)
        data.append(1)
        self.sendMessage(_cj.LM_API_CFG_LIMIT_REV, data)

        self.reverseLimit = reverseLimitPosition
        self.reverseLimitVerified = False

    def configMaxOutputVoltage(self, voltage):
        """Configure the maximum voltage that the Jaguar will ever output.

        This can be used to limit the maximum output voltage in all modes so
        that motors which cannot withstand full bus voltage can be used safely.

        :param voltage: The maximum voltage output by the Jaguar.
        """
        data = _packFXP8_8(voltage)
        self.sendMessage(_cj.LM_API_CFG_MAX_VOUT, data)

        self.maxOutputVoltage = voltage
        self.maxOutputVoltageVerified = False

    def configFaultTime(self, faultTime):
        """Configure how long the Jaguar waits in the case of a fault before
        resuming operation.

        Faults include over temerature, over current, and bus under voltage.
        The default is 3.0 seconds, but can be reduced to as low as 0.5
        seconds.

        :param faultTime: The time to wait before resuming operation, in
            seconds.
        """
        if faultTime < 0.5:
            faultTime = 0.5
        elif faultTime > 3.0:
            faultTime = 3.0

        data = _packINT16(int(faultTime * 1000.0))
        self.sendMessage(_cj.LM_API_CFG_FAULT_TIME, data)

        self.faultTime = faultTime
        self.faultTimeVerified = False

    def sendMessage(self, messageID, data,
                    period=frccan.CAN_SEND_PERIOD_NO_REPEAT):
        """Send a message to the Jaguar.

        :param messageID: The messageID to be used on the CAN bus (device
            number is added internally)
        :param data: The up to 8 bytes of data to be sent with the message
        :param period: If positive, tell Network Communications to send the
            message every "period" milliseconds.
        """
        _sendMessageHelper(messageID | self.deviceNumber, data, period)

    def requestMessage(self, messageID,
                       period=frccan.CAN_SEND_PERIOD_NO_REPEAT):
        """Request a message from the Jaguar, but don't wait for it to arrive.

        :param messageID: The message to request
        :param periodic: If positive, tell Network Communications to request
            the message every "period" milliseconds.
        """
        _sendMessageHelper(messageID | self.deviceNumber, None, period)

    def getMessage(self, messageID, messageMask):
        """Get a previously requested message.

        Jaguar always generates a message with the same message ID when
        replying.

        :param messageID: The messageID to read from the CAN bus (device
            number is added internally)
        :returns: The up to 8 bytes of data that was received with the message
        """
        messageID |= self.deviceNumber
        messageID &= _cj.CAN_MSGID_FULL_M

        # Get the data.
        messageID, data, timeStamp = \
                frccan.CANSessionMux_receiveMessage(messageID, messageMask)

        return data

    def setupPeriodicStatus(self):
        """Enables periodic status updates from the Jaguar
        """
        # Message 0 returns bus voltage, output voltage, output current, and
        # temperature.
        kMessage0Data = [
            _cj.LM_PSTAT_VOLTBUS_B0, _cj.LM_PSTAT_VOLTBUS_B1,
            _cj.LM_PSTAT_VOLTOUT_B0, _cj.LM_PSTAT_VOLTOUT_B1,
            _cj.LM_PSTAT_CURRENT_B0, _cj.LM_PSTAT_CURRENT_B1,
            _cj.LM_PSTAT_TEMP_B0, _cj.LM_PSTAT_TEMP_B1]

        # Message 1 returns position and speed
        kMessage1Data = [
            _cj.LM_PSTAT_POS_B0, _cj.LM_PSTAT_POS_B1,
            _cj.LM_PSTAT_POS_B2, _cj.LM_PSTAT_POS_B3,
            _cj.LM_PSTAT_SPD_B0, _cj.LM_PSTAT_SPD_B1,
            _cj.LM_PSTAT_SPD_B2, _cj.LM_PSTAT_SPD_B3]

        # Message 2 returns limits and faults
        kMessage2Data = [
            _cj.LM_PSTAT_LIMIT_CLR,
            _cj.LM_PSTAT_FAULT,
            _cj.LM_PSTAT_END,
            0, 0, 0, 0, 0]

        data = _packINT16(int(self.kSendMessagePeriod))
        self.sendMessage(_cj.LM_API_PSTAT_PER_EN_S0, data)
        self.sendMessage(_cj.LM_API_PSTAT_PER_EN_S1, data)
        self.sendMessage(_cj.LM_API_PSTAT_PER_EN_S2, data)

        self.sendMessage(_cj.LM_API_PSTAT_CFG_S0, kMessage0Data)
        self.sendMessage(_cj.LM_API_PSTAT_CFG_S1, kMessage1Data)
        self.sendMessage(_cj.LM_API_PSTAT_CFG_S2, kMessage2Data)

    def updatePeriodicStatus(self):
        """Check for new periodic status updates and unpack them into local
        variables.
        """
        # Check if a new bus voltage/output voltage/current/temperature message
        # has arrived and unpack the values into the cached member variables
        try:
            data = self.getMessage(_cj.LM_API_PSTAT_DATA_S0,
                                   _cj.CAN_MSGID_FULL_M)

            self.busVoltage    = _unpackFXP8_8(data[0:2])
            self.outputVoltage = _unpackPercentage(data[2:4]) * self.busVoltage
            self.outputCurrent = _unpackFXP8_8(data[4:6])
            self.temperature   = _unpackFXP8_8(data[6:8])

            self.receivedStatusMessage0 = True
        except frccan.CANMessageNotFound:
            pass

        # Check if a new position/speed message has arrived and do the same
        try:
            data = self.getMessage(_cj.LM_API_PSTAT_DATA_S1,
                                   _cj.CAN_MSGID_FULL_M)

            self.position = _unpackFXP16_16(data[0:4])
            self.speed    = _unpackFXP16_16(data[4:8])

            self.receivedStatusMessage1 = True
        except frccan.CANMessageNotFound:
            pass

        # Check if a new limits/faults message has arrived and do the same
        try:
            data = self.getMessage(_cj.LM_API_PSTAT_DATA_S2,
                                   _cj.CAN_MSGID_FULL_M)
            self.limits = data[0]
            self.faults = data[1]

            self.receivedStatusMessage2 = True
        except frccan.CANMessageNotFound:
            pass

    @staticmethod
    def updateSyncGroup(syncGroup):
        """Update all the motors that have pending sets in the syncGroup.

        :param syncGroup: A bitmask of groups to generate synchronous output.
        """
        _sendMessageHelper(_cj.CAN_MSGID_API_SYNC, [syncGroup],
                           frccan.CAN_SEND_PERIOD_NO_REPEAT)

    def getDescription(self):
        return "CANJaguar ID %d" % self.deviceNumber

    def getDeviceID(self):
        return self.deviceNumber

    def stopMotor(self):
        """Common interface for stopping a motor.
        """
        self.disableControl()

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Speed Controller"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.get())

    def valueChanged(self, itable, key, value, bln):
        self.set(float(value))

    def startLiveWindowMode(self):
        self.set(0)  # Stop for safety
        super(CANJaguar, self).startLiveWindowMode()

    def stopLiveWindowMode(self):
        super(CANJaguar, self).stopLiveWindowMode()
        self.set(0)  # Stop for safety



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .interfaces import Accelerometer
from .i2c import I2C
from .sensorbase import SensorBase
from .livewindow import LiveWindow

__all__ = ["ADXL345_I2C"]

class ADXL345_I2C(SensorBase):
    """
        ADXL345 accelerometer device via i2c
    """
    
    kAddress = 0x1D
    kPowerCtlRegister = 0x2D
    kDataFormatRegister = 0x31
    kDataRegister = 0x32
    kGsPerLSB = 0.00390625

    kPowerCtl_Link = 0x20
    kPowerCtl_AutoSleep = 0x10
    kPowerCtl_Measure = 0x08
    kPowerCtl_Sleep = 0x04

    kDataFormat_SelfTest = 0x80
    kDataFormat_SPI = 0x40
    kDataFormat_IntInvert = 0x20
    kDataFormat_FullRes = 0x08
    kDataFormat_Justify = 0x04

    Range = Accelerometer.Range

    class Axes:
        kX = 0x00
        kY = 0x02
        kZ = 0x04

    def __init__(self, port, range):
        """Constructor.

        :param port: The I2C port the accelerometer is attached to.
        :param range: The range (+ or -) that the accelerometer will measure.
        """
        self.i2c = I2C(port, self.kAddress)

        # Turn on the measurements
        self.i2c.write(self.kPowerCtlRegister, self.kPowerCtl_Measure)

        self.setRange(range)

        hal.HALReport(hal.HALUsageReporting.kResourceType_ADXL345,
                      hal.HALUsageReporting.kADXL345_I2C)

        LiveWindow.addSensor("ADXL345_I2C", port, self)

    # Accelerometer interface

    def setRange(self, range):
        """Set the measuring range of the accelerometer.

        :param range: The maximum acceleration, positive or negative, that
                      the accelerometer will measure.
        :type  range: :class:`ADXL345_I2C.Range`
        """
        if range == self.Range.k2G:
            value = 0
        elif range == self.Range.k4G:
            value = 1
        elif range == self.Range.k8G:
            value = 2
        elif range == self.Range.k16G:
            value = 3
        else:
            raise ValueError("Invalid range argument '%s'" % range)

        # Specify the data format to read
        self.i2c.write(self.kDataFormatRegister, self.kDataFormat_FullRes | value)

    def getX(self):
        """Get the x axis acceleration

        :returns: The acceleration along the x axis in g-forces
        """
        return self.getAcceleration(self.Axes.kX)

    def getY(self):
        """Get the y axis acceleration

        :returns: The acceleration along the y axis in g-forces
        """
        return self.getAcceleration(self.Axes.kY)

    def getZ(self):
        """Get the z axis acceleration

        :returns: The acceleration along the z axis in g-forces
        """
        return self.getAcceleration(self.Axes.kZ)

    def getAcceleration(self, axis):
        """Get the acceleration of one axis in Gs.

        :param axis: The axis to read from.
        :returns: An object containing the acceleration measured on each axis of the ADXL345 in Gs.
        """
        data = self.i2c.read(self.kDataRegister + axis, 2)
        # Sensor is little endian... swap bytes
        rawAccel = (data[1] << 8) | data[0]
        return rawAccel * self.kGsPerLSB

    def getAccelerations(self):
        """Get the acceleration of all axes in Gs.

        :returns: X,Y,Z tuple of acceleration measured on all axes of the
                  ADXL345 in Gs.
        """
        data = self.i2c.read(self.kDataRegister, 6)

        # Sensor is little endian... swap bytes
        rawData = []
        for i in range(3):
            rawData.append((data[i*2+1] << 8) | data[i*2])

        return (rawData[0] * self.kGsPerLSB,
                rawData[1] * self.kGsPerLSB,
                rawData[2] * self.kGsPerLSB)

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "3AxisAccelerometer"

    def initTable(self, subtable):
        self.table = subtable
        self.updateTable()

    def getTable(self):
        return self.table

    def updateTable(self):
        if self.table is not None:
            self.table.putNumber("X", self.getX())
            self.table.putNumber("Y", self.getY())
            self.table.putNumber("Z", self.getZ())

    def startLiveWindowMode(self):
        """
        ADXL345_I2C doesn't have to do anything special when entering the LiveWindow.
        """
        pass

    def stopLiveWindowMode(self):
        """
        ADXL345_I2C doesn't have to do anything special when exiting the LiveWindow.
        """
        pass



"""
    Objects in this package allow you to implement a robot using Command-based
    programming.  Command based programming is a design pattern to help you
    organize your robot programs, by organizing your robot program into
    components based on :class:`.Command` and :class:`.Subsystem`
    
    The python implementation of the Command framework closely follows the 
    Java language implementation. RobotPy has several examples of command 
    based robots available.
    
    Each one of the objects in the Command framework has detailed
    documentation available. If you need more information, for examples,
    tutorials, and other detailed information on programming your robot
    using this pattern, we recommend that you consult the Java version of the
    `FRC Control System documentation <https://wpilib.screenstepslive.com/s/3120/m/7952/c/44956>`_
"""

from .command import *
from .commandgroup import *
from .pidcommand import *
from .pidsubsystem import *
from .printcommand import *
from .scheduler import *
from .startcommand import *
from .subsystem import *
from .waitcommand import *
from .waitforchildren import *
from .waituntilcommand import *



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import warnings
import weakref

from .interfaces.counterbase import CounterBase
from .interfaces.pidsource import PIDSource
from .analogtriggeroutput import AnalogTriggerOutput
from .digitalinput import DigitalInput
from .livewindow import LiveWindow
from .sensorbase import SensorBase
from ._impl.utils import match_arglist, HasAttribute

__all__ = ["Counter"]

def _freeCounter(counter):
    hal.setCounterUpdateWhenEmpty(counter, True)
    hal.clearCounterUpSource(counter)
    hal.clearCounterDownSource(counter)
    hal.freeCounter(counter)

class Counter(SensorBase):
    """Counts the number of ticks on a :class:`.DigitalInput` channel.
    
    This is a general purpose class for counting repetitive events. It can return
    the number of counts, the period of the most recent cycle, and detect when
    the signal being counted has stopped by supplying a maximum cycle time.

    All counters will immediately start counting - :meth:`reset` them if you need
    them to be zeroed before use.
    
    .. not_implemented: initCounter
    """

    class Mode:
        """Mode determines how and what the counter counts"""
        
        #: two pulse mode
        kTwoPulse = 0
        
        #: semi period mode
        kSemiperiod = 1
        
        #: pulse length mode
        kPulseLength = 2
        
        #: external direction mode
        kExternalDirection = 3

    EncodingType = CounterBase.EncodingType
    PIDSourceParameter = PIDSource.PIDSourceParameter
    allocatedUpSource = False
    allocatedDownSource = False

    def __init__(self, *args, **kwargs):
        """Counter constructor.

        The counter will start counting immediately.

        Positional arguments may be either channel numbers, :class:`.DigitalSource`
        sources, or :class:`.AnalogTrigger` sources in the following order:

        A "source" is any valid single-argument input to :meth:`setUpSource` and :meth:`setDownSource`
        
        - (none)
        - upSource
        - upSource, down source
        And, to keep consistency with Java wpilib.
        - encodingType, up source, down source, inverted

        If the passed object has a
        `getChannelForRouting` function, it is assumed to be a DigitalSource.
        If the passed object has a `createOutput` function, it is assumed to
        be an AnalogTrigger.

        In addition, extra keyword parameters may be provided for mode, inverted,
        and encodingType.

        :param upSource: The source (channel num, DigitalInput, or AnalogTrigger)
            that should be used for up counting.
        :param downSource: The source (channel num, DigitalInput, or AnalogTrigger)
            that should be used for down counting or direction control.
        :param mode:
            How and what the counter counts (see :class:`.Mode`).  Defaults to
            `Mode.kTwoPulse` for zero or one source, and
            `Mode.kExternalDirection` for two sources.
        :param inverted:
            Flips the direction of counting.  Defaults to False if unspecified.
            Only used when two sources are specified.
        :param encodingType:
            Either k1X or k2X to indicate 1X or 2X decoding. 4X decoding
            is not supported by Counter; use `Encoder` instead.  Defaults
            to k1X if unspecified.  Only used when two sources are specified.
        """

        source_identifier = [int, HasAttribute("getChannelForRouting"), HasAttribute("createOutput")]

        argument_templates = [[],
                              [("upSource", source_identifier), ],
                              [("upSource", source_identifier), ("downSource", source_identifier)],
                              [("encodingType", None), ("upSource", source_identifier),
                               ("downSource", source_identifier), ("inverted", bool)], ]


        _, results = match_arglist('Counter.__init__',
                                   args, kwargs, argument_templates, allow_extra_kwargs=True)

        # extract arguments
        upSource = results.pop("upSource", None)
        downSource = results.pop("downSource", None)

        encodingType = results.pop("encodingType", None)
        inverted = results.pop("inverted", False)
        mode = results.pop("mode", None)

        if mode is None:
            #Get the mode
            if upSource is not None and downSource is not None:
                mode = self.Mode.kExternalDirection
            else:
                mode = self.Mode.kTwoPulse

        # save some variables
        self.distancePerPulse = 1.0 # distance of travel for each tick
        self.pidSource = PIDSource.PIDSourceParameter.kDistance

        # create counter
        self._counter, self.index = hal.initializeCounter(mode)
        self._counter_finalizer = \
            weakref.finalize(self, _freeCounter, self._counter)

        self.setMaxPeriod(.5)

        hal.HALReport(hal.HALUsageReporting.kResourceType_Counter, self.index,
                      mode)

        #Set sources
        if upSource is not None:
            self.setUpSource(upSource)
        else:
            self.upSource = None

        if downSource is not None:
            self.setDownSource(downSource)
        else:
            self.downSource = None

        # when given two sources, set edges
        if upSource is not None and downSource is not None:
            if encodingType == self.EncodingType.k1X:
                self.setUpSourceEdge(True, False)
                hal.setCounterAverageSize(self._counter, 1)
            else:
                self.setUpSourceEdge(True, True)
                hal.setCounterAverageSize(self._counter, 2)
            self.setDownSourceEdge(inverted, True)

    @property
    def counter(self):
        if not self._counter_finalizer.alive:
            return None
        return self._counter

    def free(self):
        self.setUpdateWhenEmpty(True)
        self.clearUpSource()

        self.clearDownSource()
        self._counter_finalizer()

    def getFPGAIndex(self):
        """
        :returns: The Counter's FPGA index.
        """
        return self.index

    def setUpSource(self, *args, **kwargs):
        """Set the up counting source for the counter.

        This function accepts either a digital channel index, a
        `DigitalSource`, or an `AnalogTrigger` as positional arguments:

        - source
        - channel
        - analogTrigger
        - analogTrigger, triggerType

        For positional arguments, if the passed object has a
        `getChannelForRouting` function, it is assumed to be a DigitalSource.
        If the passed object has a `createOutput` function, it is assumed to
        be an AnalogTrigger.

        Alternatively, sources and/or channels may be passed as keyword
        arguments.  The behavior of specifying both a source and a number
        for the same channel is undefined, as is passing both a positional
        and a keyword argument for the same channel.

        :param channel: the DIO channel to use as the up source. 0-9 are on-board, 10-25 are on the MXP
        :type channel: int
        :param source: The digital source to count
        :type source: DigitalInput
        :param analogTrigger:
            The analog trigger object that is used for the Up Source
        :type analogTrigger: AnalogTrigger
        :param triggerType:
            The analog trigger output that will trigger the counter.
            Defaults to kState if not specified.
        :type triggerType: AnalogTriggerType
        """

        #TODO Both this and the java implementation should probably not allow setting a source if one is already set.

        if self.counter is None:
            raise ValueError("operation on freed port")

        argument_templates = [[("channel", int)],
                              [("source", HasAttribute("getChannelForRouting")), ],
                              [("analogTrigger", HasAttribute("createOutput"))],
                              [("analogTrigger", HasAttribute("createOutput")), ("triggerType", None)]]

        _, results = match_arglist('Counter.setUpSource',
                                   args, kwargs, argument_templates)

        # extract arguments
        source = results.pop("source", None)
        channel = results.pop("channel", None)
        analogTrigger = results.pop("analogTrigger", None)
        triggerType = results.pop("triggerType", AnalogTriggerOutput.AnalogTriggerType.kState)

        # If we don't have source, generate it from other arguments.
        if source is None:
            if channel is not None:
                source = DigitalInput(channel)
                self.allocatedUpSource = True
            elif analogTrigger is not None and triggerType is not None:
                source = analogTrigger.createOutput(triggerType)
            else:
                raise ValueError("No usable source.")

        # save and set
        self.upSource = source
        hal.setCounterUpSource(self.counter,
                               self.upSource.getChannelForRouting(),
                               self.upSource.getAnalogTriggerForRouting())

    def setUpSourceEdge(self, risingEdge, fallingEdge):
        """Set the edge sensitivity on an up counting source. Set the up
        source to either detect rising edges or falling edges.

        :param risingEdge: True to count rising edge
        :type  risingEdge: bool
        :param fallingEdge: True to count falling edge
        :type  fallingEdge: bool
        """
        if self.upSource is None:
            raise ValueError("Up Source must be set before setting the edge")
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterUpSourceEdge(self.counter, risingEdge, fallingEdge)

    def clearUpSource(self):
        """Disable the up counting source to the counter."""
        if self.upSource is not None and self.allocatedUpSource:
            self.upSource.free()
            self.allocatedUpSource = False
        self.upSource = None

        if self.counter is None:
            return
        hal.clearCounterUpSource(self.counter)

    def setDownSource(self, *args, **kwargs):
        """Set the down counting source for the counter.

        This function accepts either a digital channel index, a
        `DigitalSource`, or an `AnalogTrigger` as positional arguments:
        
        - source
        - channel
        - analogTrigger
        - analogTrigger, triggerType

        For positional arguments, if the passed object has a
        `getChannelForRouting` function, it is assumed to be a DigitalSource.
        If the passed object has a `createOutput` function, it is assumed to
        be an AnalogTrigger.

        Alternatively, sources and/or channels may be passed as keyword
        arguments.  The behavior of specifying both a source and a number
        for the same channel is undefined, as is passing both a positional
        and a keyword argument for the same channel.

        :param channel: the DIO channel to use as the down source. 0-9 are on-board, 10-25 are on the MXP
        :type channel: int
        :param source: The digital source to count
        :type source: DigitalInput
        :param analogTrigger:
            The analog trigger object that is used for the Up Source
        :type analogTrigger: AnalogTrigger
        :param triggerType:
            The analog trigger output that will trigger the counter.
            Defaults to kState if not specified.
        :type triggerType: AnalogTriggerType
        """

        #TODO Both this and the java implementation should probably not allow setting a source if one is already set.

        if self.counter is None:
            raise ValueError("operation on freed port")

        argument_templates = [[("channel", int)],
                              [("source", HasAttribute("getChannelForRouting")), ],
                              [("analogTrigger", HasAttribute("createOutput")), ],
                              [("analogTrigger", HasAttribute("createOutput")), ("triggerType", None)]]

        _, results = match_arglist('Counter.setUpSource',
                                   args, kwargs, argument_templates)

        # extract arguments
        source = results.pop("source", None)
        channel = results.pop("channel", None)
        analogTrigger = results.pop("analogTrigger", None)
        triggerType = results.pop("triggerType", AnalogTriggerOutput.AnalogTriggerType.kState)

        # If we don't have source, generate it from other arguments.
        if source is None:
            if channel is not None:
                source = DigitalInput(channel)
                self.allocatedDownSource = True
            elif analogTrigger is not None and triggerType is not None:
                source = analogTrigger.createOutput(triggerType)
            else:
                raise ValueError("No usable source.")

        # save and set
        self.downSource = source
        hal.setCounterDownSource(self.counter,
                                 self.downSource.getChannelForRouting(),
                                 self.downSource.getAnalogTriggerForRouting())

    def setDownSourceEdge(self, risingEdge, fallingEdge):
        """Set the edge sensitivity on an down counting source. Set the down
        source to either detect rising edges or falling edges.

        :param risingEdge: True to count rising edge
        :type  risingEdge: bool
        :param fallingEdge: True to count falling edge
        :type  fallingEdge: bool
        """
        if self.downSource is None:
            raise ValueError("Down Source must be set before setting the edge")
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterDownSourceEdge(self.counter, risingEdge, fallingEdge)

    def clearDownSource(self):
        """Disable the down counting source to the counter.
        """
        if self.downSource is not None and self.allocatedDownSource:
            self.downSource.free()
            self.allocatedDownSource = False
        self.downSource = None

        if self.counter is None:
            return
        hal.clearCounterDownSource(self.counter)

    def setUpDownCounterMode(self):
        """Set standard up / down counting mode on this counter. Up and down
        counts are sourced independently from two inputs.
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterUpDownMode(self.counter)

    def setExternalDirectionMode(self):
        """Set external direction mode on this counter. Counts are sourced on
        the Up counter input. The Down counter input represents the direction
        to count.
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterExternalDirectionMode(self.counter)

    def setSemiPeriodMode(self, highSemiPeriod):
        """Set Semi-period mode on this counter. Counts up on both rising and
        falling edges.

        :param highSemiPeriod: True to count up on both rising and falling
        :type  highSemiPeriod: bool
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterSemiPeriodMode(self.counter, highSemiPeriod)

    def setPulseLengthMode(self, threshold):
        """Configure the counter to count in up or down based on the length
        of the input pulse. This mode is most useful for direction sensitive
        gear tooth sensors.

        :param threshold: The pulse length beyond which the counter counts the
            opposite direction. Units are seconds.
        :type  threshold: float, int
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterPulseLengthMode(self.counter, float(threshold))

    def get(self):
        """Read the current counter value. Read the value at this instant. It
        may still be running, so it reflects the current value. Next time it
        is read, it might have a different value.
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        return hal.getCounter(self.counter)

    def getDistance(self):
        """Read the current scaled counter value. Read the value at this
        instant, scaled by the distance per pulse (defaults to 1).

        :returns: Scaled value
        :rtype: float
        """
        return self.get() * self.distancePerPulse

    def reset(self):
        """Reset the Counter to zero. Set the counter value to zero. This
        doesn't effect the running state of the counter, just sets the
        current value to zero.
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.resetCounter(self.counter)

    def setMaxPeriod(self, maxPeriod):
        """Set the maximum period where the device is still considered
        "moving".  Sets the maximum period where the device is considered
        moving. This value is used to determine the "stopped" state of the
        counter using the :func:`getStopped` method.

        :param maxPeriod: The maximum period where the counted device is
            considered moving in seconds.
        :type maxPeriod: float or int
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterMaxPeriod(self.counter, float(maxPeriod))

    def setUpdateWhenEmpty(self, enabled):
        """Select whether you want to continue updating the event timer
        output when there are no samples captured. The output of the event
        timer has a buffer of periods that are averaged and posted to a
        register on the FPGA. When the timer detects that the event source
        has stopped (based on the MaxPeriod) the buffer of samples to be
        averaged is emptied. If you enable update when empty, you will be
        notified of the stopped source and the event time will report 0
        samples. If you disable update when empty, the most recent average
        will remain on the output until a new sample is acquired. You will
        never see 0 samples output (except when there have been no events
        since an FPGA reset) and you will likely not see the stopped bit
        become true (since it is updated at the end of an average and
        there are no samples to average).

        :param enabled: True to continue updating
        :type  enabled: bool
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterUpdateWhenEmpty(self.counter, enabled)

    def getStopped(self):
        """Determine if the clock is stopped. Determine if the clocked input
        is stopped based on the MaxPeriod value set using the
        :func:`setMaxPeriod` method.  If the clock exceeds the MaxPeriod,
        then the device (and counter) are assumed to be stopped and it
        returns True.

        :returns: Returns True if the most recent counter period exceeds the
            MaxPeriod value set by SetMaxPeriod.
        :rtype: bool
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        return hal.getCounterStopped(self.counter)

    def getDirection(self):
        """The last direction the counter value changed.

        :returns: The last direction the counter value changed.
        :rtype: bool
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        return hal.getCounterDirection(self.counter)

    def setReverseDirection(self, reverseDirection):
        """Set the Counter to return reversed sensing on the direction. This
        allows counters to change the direction they are counting in the case
        of 1X and 2X quadrature encoding only. Any other counter mode isn't
        supported.

        :param reverseDirection: True if the value counted should be negated.
        :type  reverseDirection: bool 
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterReverseDirection(self.counter, reverseDirection)

    def getPeriod(self):
        """Get the Period of the most recent count. Returns the time interval
        of the most recent count. This can be used for velocity calculations
        to determine shaft speed.

        :returns: The period of the last two pulses in units of seconds.
        :rtype: float
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        return hal.getCounterPeriod(self.counter)

    def getRate(self):
        """Get the current rate of the Counter. Read the current rate of the
        counter accounting for the distance per pulse value. The default
        value for distance per pulse (1) yields units of pulses per second.

        :returns: The rate in units/sec
        :rtype: float
        """
        return self.distancePerPulse / self.getPeriod()

    def setSamplesToAverage(self, samplesToAverage):
        """Set the Samples to Average which specifies the number of samples
        of the timer to average when calculating the period. Perform averaging
        to account for mechanical imperfections or as oversampling to increase
        resolution.

        :param samplesToAverage: The number of samples to average from 1 to 127.
        :type  samplesToAverage: int
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        hal.setCounterSamplesToAverage(self.counter, samplesToAverage)

    def getSamplesToAverage(self):
        """Get the Samples to Average which specifies the number of samples
        of the timer to average when calculating the period. Perform averaging
        to account for mechanical imperfections or as oversampling to increase
        resolution.

        :returns: The number of samples being averaged (from 1 to 127)
        :rtype: int
        """
        if self.counter is None:
            raise ValueError("operation on freed port")
        return hal.getCounterSamplesToAverage(self.counter)

    def setDistancePerPulse(self, distancePerPulse):
        """Set the distance per pulse for this counter. This sets the
        multiplier used to determine the distance driven based on the count
        value from the encoder. Set this value based on the Pulses per
        Revolution and factor in any gearing reductions. This distance can be
        in any units you like, linear or angular.

        :param distancePerPulse:
            The scale factor that will be used to convert pulses to useful
            units.
        :type  distancePerPulse: float
        """
        self.distancePerPulse = distancePerPulse

    def setPIDSourceParameter(self, pidSource):
        """Set which parameter of the encoder you are using as a process
        control variable. The counter class supports the rate and distance
        parameters.

        :param pidSource: An enum to select the parameter.
        :type  pidSource: :class:`Counter.PIDSourceParameter`
        """
        if pidSource not in (self.PIDSourceParameter.kDistance,
                             self.PIDSourceParameter.kRate):
            raise ValueError("Invalid pidSource argument '%s'" % pidSource)
        self.pidSource = pidSource

    def pidGet(self):
        if self.pidSource == self.PIDSourceParameter.kDistance:
            return self.getDistance()
        elif self.pidSource == self.PIDSourceParameter.kRate:
            return self.getRate()
        else:
            return 0.0

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Counter"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.get())

    def startLiveWindowMode(self):
        pass

    def stopLiveWindowMode(self):
        pass



'''
    An implementation of java.util.TimerTask which is needed for
    the PID Controller
'''

import threading
from ..timer import Timer


class TimerTask(threading.Thread):
    
    def __init__(self, name, period, task_fn):
        super().__init__(name=name, daemon=True)
        self.period = period
        self.task_fn = task_fn
        
        self.stopped = False
    
    def cancel(self):
        self.stopped = True
        self.join()
    
    def run(self):
        
        period = self.period
        wait_til = Timer.getFPGATimestamp() + period
        
        while not self.stopped:
            
            now = Timer.getFPGATimestamp()
            Timer.delay(wait_til - now)
            
            if self.stopped:
                break
            
            self.task_fn()
            
            wait_til += period
    


#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

__all__ = ["PIDSource"]

class PIDSource:
    """This interface allows for :class:`.PIDController` to automatically read from this
    object.
    """

    class PIDSourceParameter:
        """A description for the type of output value to provide to a
        :class:`.PIDController`"""
        kDistance = 0
        kRate = 1
        kAngle = 2

    def pidGet(self):
        """Get the result to use in :class:`.PIDController`
        
        :returns: the result to use in PIDController
        """
        raise NotImplementedError



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .button import Button

__all__ = ["InternalButton"]

class InternalButton(Button):
    """This class is intended to be used within a program.  The programmer can
    manually set its value. Includes a setting for whether or not it should
    invert its value.
    """

    def __init__(self, inverted=False):
        """Creates an InternalButton which is inverted depending on the input.

        :param inverted: If False, then this button is pressed when set to
                         True, otherwise it is pressed when set to False.
        """
        self.pressed = inverted
        self.inverted = inverted

    def setInverted(self, inverted):
        self.inverted = inverted

    def setPressed(self, pressed):
        self.pressed = pressed

    def get(self):
        return self.pressed ^ self.inverted



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

__all__ = ["GenericHID"]

class GenericHID:
    """GenericHID Interface"""

    class Hand:
        """Which hand the Human Interface Device is associated with."""
        
        #: Left hand
        kLeft = 0
        
        #: Right hand
        kRight = 1

    def getX(self, hand=None):
        """Get the x position of HID.

        :param hand: which hand, left or right (default right)
        :returns: the x position
        """
        raise NotImplementedError

    def getY(self, hand=None):
        """Get the y position of the HID.

        :param hand: which hand, left or right (default right)
        :returns: the y position
        """
        raise NotImplementedError

    def getZ(self, hand=None):
        """Get the z position of the HID.

        :param hand: which hand, left or right (default right)
        :returns: the z position
        """
        raise NotImplementedError

    def getTwist(self):
        """Get the twist value.

        :returns: the twist value
        """
        raise NotImplementedError

    def getThrottle(self):
        """Get the throttle.

        :returns: the throttle value
        """
        raise NotImplementedError

    def getRawAxis(self, which):
        """Get the raw axis.

        :param which: index of the axis
        :returns: the raw value of the selected axis
        """
        raise NotImplementedError

    def getTrigger(self, hand=None):
        """Is the trigger pressed
        
        :param hand: which hand (default right)
        :returns: True if the trigger for the given hand is pressed
        """
        raise NotImplementedError

    def getTop(self, hand=None):
        """Is the top button pressed
        
        :param hand: which hand (default right)
        :returns: True if the top button for the given hand is pressed
        """
        raise NotImplementedError

    def getBumper(self, hand=None):
        """Is the bumper pressed?

        :param hand: which hand (default right)
        :returns: True if the bumper is pressed
        """
        raise NotImplementedError

    def getRawButton(self, button):
        """Is the given button pressed?

        :param button: which button number
        :returns: True if the button is pressed
        """
        raise NotImplementedError

    def getPOV(self, pov=0):
        """Get the state of a POV.

        :param pov: which POV (default is 0)
        :returns: The angle of the POV in degrees, or -1 if the POV is not
                  pressed.
        """
        raise NotImplementedError



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase

__all__ = ["AnalogOutput"]

class AnalogOutput(SensorBase):
    """Analog output"""

    channels = Resource(SensorBase.kAnalogOutputChannels)

    def __init__(self, channel):
        """Construct an analog output on a specified MXP channel.

        :param channel: The channel number to represent.
        """
        if not hal.checkAnalogOutputChannel(channel):
            raise IndexError("Analog output channel %d cannot be allocated. Channel is not present." % channel)
        try:
            AnalogOutput.channels.allocate(self, channel)
        except IndexError as e:
            raise IndexError("Analog output channel %d is already allocated" % channel) from e

        self.channel = channel

        port = hal.getPort(channel)
        self.port = hal.initializeAnalogOutputPort(port)

        LiveWindow.addSensorChannel("AnalogOutput", channel, self)
        hal.HALReport(hal.HALUsageReporting.kResourceType_AnalogChannel,
                      channel, 1)

    def free(self):
        """Channel destructor.
        """
        if self.channel is None:
            return
        AnalogOutput.channels.free(self.channel)
        self.channel = None

    def setVoltage(self, voltage):
        hal.setAnalogOutput(self.port, voltage)

    def getVoltage(self):
        return hal.getAnalogOutput(self.port)

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Analog Output"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.getVoltage())

    def startLiveWindowMode(self):
        # Analog Channels don't have to do anything special when entering the
        # LiveWindow.
        pass

    def stopLiveWindowMode(self):
        # Analog Channels don't have to do anything special when exiting the
        # LiveWindow.
        pass



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

from .resource import Resource
from .sensorbase import SensorBase
from .interruptablesensorbase import InterruptableSensorBase

__all__ = ["DigitalSource"]

class DigitalSource(InterruptableSensorBase):
    """DigitalSource Interface. The DigitalSource represents all the possible
    inputs for a counter or a quadrature encoder. The source may be either a
    digital input or an analog input. If the caller just provides a channel,
    then a digital input will be constructed and freed when finished for the
    source. The source can either be a digital input or analog trigger but
    not both.
    """

    channels = Resource(SensorBase.kDigitalChannels)

    def __init__(self, channel, input):
        """
            :param channel: Port for the digital input
            :type  channel: int
            :param input: True if input, False otherwise
            :type  input: int
        """
        super().__init__()

        self.channel = channel

        # XXX: Replace with hal.checkDigitalChannel when implemented
        SensorBase.checkDigitalChannel(channel)

        try:
            DigitalSource.channels.allocate(self, channel)
        except IndexError as e:
            raise IndexError("Digital input %d is already allocated" % self.channel) from e

        self._port = hal.initializeDigitalPort(hal.getPort(channel))
        hal.allocateDIO(self._port, True if input else False)

        self._port_finalizer = weakref.finalize(self, hal.freeDIO, self._port)

    @property
    def port(self):
        if not self._port_finalizer.alive:
            return None
        return self._port

    def free(self):
        if self.channel is None:
            return
        DigitalSource.channels.free(self.channel)
        self._port_finalizer()
        self.channel = None

    def getChannelForRouting(self):
        """Get the channel routing number

        :returns: channel routing number
        """
        return self.channel

    def getModuleForRouting(self):
        """Get the module routing number

        :returns: 0
        """
        return 0

    def getAnalogTriggerForRouting(self):
        """Is this an analog trigger
        
        :returns: True if this is an analog trigger
        """
        return False



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .button import Button

__all__ = ["JoystickButton"]

class JoystickButton(Button):
    def __init__(self, joystick, buttonNumber):
        """Create a joystick button for triggering commands.

        :param joystick: The GenericHID object that has the button (e.g.
                         :class:`.Joystick`, :class:`.KinectStick`, etc)
        :param buttonNumber: The button number
                             (see :meth:`GenericHID.getRawButton`)
        """
        super().__init__()
        self.joystick = joystick
        self.buttonNumber = buttonNumber

    def get(self):
        """Gets the value of the joystick button.

        :returns: The value of the joystick button
        """
        return self.joystick.getRawButton(self.buttonNumber)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .command import Command

from ..timer import Timer

__all__ = ["WaitUntilCommand"]

class WaitUntilCommand(Command):
    """
    This will wait until the game clock reaches some value, then continue to
    the next command.
    """

    def __init__(self, time):
        super().__init__("WaitUntil(%s)" % time)
        self.time = time

    def isFinished(self):
        # Check if we've reached the actual finish time.
        return Timer.getMatchTime() >= self.time



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import warnings

from .livewindow import LiveWindow
from .sensorbase import SensorBase
from .solenoidbase import SolenoidBase

__all__ = ["Solenoid"]

class Solenoid(SolenoidBase):
    """Solenoid class for running high voltage Digital Output.

    The Solenoid class is typically used for pneumatics solenoids, but could
    be used for any device within the current spec of the PCM.
    
    .. not_implemented: initSolenoid
    """

    def __init__(self, *args, **kwargs):
        """Constructor.

        Arguments can be supplied as positional or keyword.  Acceptable
        positional argument combinations are:
        
        - channel
        - moduleNumber, channel

        Alternatively, the above names can be used as keyword arguments.

        :param moduleNumber: The CAN ID of the PCM the solenoid is attached to
        :type moduleNumber: int
        :param channel: The channel on the PCM to control (0..7)
        :type channel: int
        """
        # keyword arguments
        channel = kwargs.pop("channel", None)
        moduleNumber = kwargs.pop("moduleNumber", None)

        if kwargs:
            warnings.warn("unknown keyword arguments: %s" % kwargs.keys(),
                          RuntimeWarning)

        # positional arguments
        if len(args) == 1:
            channel = args[0]
        elif len(args) == 2:
            moduleNumber, channel = args
        elif len(args) != 0:
            raise ValueError("don't know how to handle %d positional arguments" % len(args))

        if moduleNumber is None:
            moduleNumber = SensorBase.getDefaultSolenoidModule()
        if channel is None:
            raise ValueError("must specify channel")

        SensorBase.checkSolenoidModule(moduleNumber)
        SensorBase.checkSolenoidChannel(channel)

        super().__init__(moduleNumber)
        self.channel = channel

        try:
            self.allocated.allocate(self, channel)
        except IndexError as e:
            raise IndexError("Solenoid channel %d on module %d is already allocated" % (channel, moduleNumber)) from e

        self.port = self.ports[channel]

        LiveWindow.addActuatorModuleChannel("Solenoid", moduleNumber, channel,
                                            self)
        hal.HALReport(hal.HALUsageReporting.kResourceType_Solenoid, channel,
                      moduleNumber)

    def free(self):
        """Mark the solenoid as freed."""
        self.allocated.free(self.channel)

    def set(self, on):
        """Set the value of a solenoid.

        :param on: Turn the solenoid output off or on.
        :type on: bool
        """
        with self.mutex:
            hal.setSolenoid(self.port, on)

    def get(self):
        """Read the current value of the solenoid.

        :returns: The current value of the solenoid.
        :rtype: bool
        """
        with self.mutex:
            return hal.getSolenoid(self.port)

    def isBlackListed(self):
        """
        Check if the solenoid is blacklisted.
            If a solenoid is shorted, it is added to the blacklist and disabled until power cycle, or until faults are
            cleared. See :meth:`clearAllPCMStickyFaults`

        :returns: If solenoid is disabled due to short.
        """
        value = self.getPCMSolenoidBlackList() & (1 << self.channel)
        return value != 0

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Solenoid"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putBoolean("Value", self.get())

    def valueChanged(self, itable, key, value, bln):
        self.set(True if value else False)

    def startLiveWindowMode(self):
        self.set(False) # Stop for safety
        super().startLiveWindowMode()

    def stopLiveWindowMode(self):
        super().stopLiveWindowMode()
        self.set(False) # Stop for safety



import hal
import weakref

from .livewindow import LiveWindow
from .livewindowsendable import LiveWindowSendable
from .motorsafety import MotorSafety
from .resource import Resource
from .timer import Timer

__all__ = ["CANTalon"]

def _freeCANTalon(handle):
    hal.TalonSRX_Destroy(handle)

class CANTalon(LiveWindowSendable, MotorSafety):
    """Talon SRX device as a CAN device

    The TALON SRX is designed to instrument all runtime signals periodically.
    The default periods are chosen to support 16 TALONs with 10ms update rate
    for control (throttle or setpoint).  However these can be overridden with
    :meth:`setStatusFrameRate`.

    Likewise most control signals are sent periodically using the
    fire-and-forget CAN API.

    Signals that are not available in an unsolicited fashion are the Close
    Loop gains.  For teams that have a single profile for their TALON close
    loop they can use either the webpage to configure their TALONs once or
    set the PIDF,Izone,CloseLoopRampRate,etc... once in the robot application.
    These parameters are saved to flash so once they are loaded in the TALON,
    they will persist through power cycles and mode changes.

    For teams that have one or two profiles to switch between, they can use
    the same strategy since there are two slots to choose from and the
    ProfileSlotSelect is periodically sent in the 10 ms control frame.

    For teams that require changing gains frequently, they can use the
    soliciting API to get and set those parameters.  Most likely they will
    only need to set them in a periodic fashion as a function of what motion
    the application is attempting.  If this API is used, be mindful of the CAN
    utilization reported in the driver station.

    Encoder position is measured in encoder edges.  Every edge is counted
    (similar to roboRIO 4X mode).  Analog position is 10 bits, meaning 1024
    ticks per rotation (0V => 3.3V).  Use :meth:`setFeedbackDevice` to select
    which sensor type you need.  Once you do that you can use
    :meth:`getSensorPosition` and :meth:`getSensorVelocity`.  These signals
    are updated on CANBus every 20ms (by default).  If a relative sensor is
    selected, you can zero (or change the current value) using
    :meth:`setSensorPosition`.

    Analog Input and quadrature position (and velocity) are also explicitly
    reported in :meth:`getEncPosition`, :meth:`getEncVelocity`,
    :meth:`getAnalogInPosition`, :meth:`getAnalogInRaw`,
    :meth:`getAnalogInVelocity`.
    These signals are available all the time, regardless of what sensor is
    selected at a rate of 100ms.  This allows easy instrumentation for "in the
    pits" checking of all sensors regardless of modeselect.  The 100ms rate is
    overridable for teams who want to acquire sensor data for processing, not
    just instrumentation.  Or just select the sensor using
    :meth:`setFeedbackDevice` to get it at 20ms.

    Velocity is in position ticks / 100ms.

    All output units are in respect to duty cycle (throttle) which is
    -1023(full reverse) to +1023 (full forward).  This includes demand (which
    specifies duty cycle when in duty cycle mode) and rampRamp, which is in
    throttle units per 10ms (if nonzero).

    When in (default) PercentVBus mode, set() and get() are automatically
    scaled to a -1.0 to +1.0 range to match other motor controllers.

    Pos and velocity close loops are calc'd as::

        err = target - posOrVel
        iErr += err
        if IZone != 0 and abs(err) > IZone:
            ClearIaccum()
        output = P * err + I * iErr + D * dErr + F * target
        dErr = err - lastErr

    P, I, and D gains are always positive. F can be negative.

    Motor direction can be reversed using :meth:`reverseOutput` if sensor
    and motor are out of phase.  Similarly feedback sensor can also be reversed
    (multiplied by -1) using :meth:`reverseSensor` if you prefer the sensor to
    be inverted.

    P gain is specified in throttle per error tick.  For example, a value of
    102 is ~9.9% (which is 102/1023) throttle per 1 ADC unit(10bit) or 1
    quadrature encoder edge depending on selected sensor.

    I gain is specified in throttle per integrated error. For example, a value
    of 10 equates to ~0.99% (which is 10/1023) for each accumulated ADC
    unit(10bit) or 1 quadrature encoder edge depending on selected sensor.
    Close loop and integral accumulator runs every 1ms.

    D gain is specified in throttle per derivative error. For example a value
    of 102 equates to ~9.9% (which is 102/1023) per change of 1 unit (ADC or
    encoder) per ms.

    I Zone is specified in the same units as sensor position (ADC units or
    quadrature edges).  If pos/vel error is outside of this value, the
    integrated error will auto-clear::

        if IZone != 0 and abs(err) > IZone:
            ClearIaccum()

    This is very useful in preventing integral windup and is highly
    recommended if using full PID to keep stability low.

    CloseLoopRampRate is in throttle units per 1ms.  Set to zero to disable
    ramping.  Works the same as RampThrottle but only is in effect when a
    close loop mode and profile slot is selected.
    """

    class ControlMode:
        PercentVbus = hal.TalonSRXConst.kMode_DutyCycle
        Position = hal.TalonSRXConst.kMode_PositionCloseLoop
        Speed = hal.TalonSRXConst.kMode_VelocityCloseLoop
        Current = hal.TalonSRXConst.kMode_CurrentCloseLoop
        Voltage = hal.TalonSRXConst.kMode_VoltCompen
        Follower = hal.TalonSRXConst.kMode_SlaveFollower
        Disabled = hal.TalonSRXConst.kMode_NoDrive

    class FeedbackDevice:
        QuadEncoder = hal.TalonSRXConst.kFeedbackDev_DigitalQuadEnc
        AnalogPot = hal.TalonSRXConst.kFeedbackDev_AnalogPot
        AnalogEncoder = hal.TalonSRXConst.kFeedbackDev_AnalogEncoder
        EncRising = hal.TalonSRXConst.kFeedbackDev_CountEveryRisingEdge
        EncFalling = hal.TalonSRXConst.kFeedbackDev_CountEveryFallingEdge

    class StatusFrameRate:
        """enumerated types for frame rate ms"""
        General = hal.TalonSRXConst.kStatusFrame_General
        Feedback = hal.TalonSRXConst.kStatusFrame_Feedback
        QuadEncoder = hal.TalonSRXConst.kStatusFrame_Encoder
        AnalogTempVbat = hal.TalonSRXConst.kStatusFrame_AnalogTempVbat

    kDelayForSolicitedSignals = 0.004

    def __init__(self, deviceNumber,
                 controlPeriodMs=hal.TalonSRXConst.kDefaultControlPeriodMs):
        MotorSafety.__init__(self)

        self.deviceNumber = deviceNumber
        # HAL bounds period to be within [1 ms,95 ms]
        self._handle = hal.TalonSRX_Create(deviceNumber, controlPeriodMs)
        self._handle_finalizer = weakref.finalize(self, _freeCANTalon,
                                                  self._handle)
        self.controlEnabled = True
        self.profile = 0
        self.setPoint = 0.0
        self.setProfile(self.profile)
        self._applyControlMode(self.ControlMode.PercentVbus)
        
        LiveWindow.addActuatorChannel("CANTalonSRX", deviceNumber, self)
        hal.HALReport(hal.HALUsageReporting.kResourceType_CANTalonSRX,
                      deviceNumber)
        
        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

    @property
    def handle(self):
        if not self._handle_finalizer.alive:
            raise ValueError("operation on freed port")
        return self._handle

    def free(self):
        self._handle_finalizer()

    def pidWrite(self, output):
        if self.getControlMode() == self.ControlMode.PercentVbus:
            self.set(output)
        else:
            raise ValueError("PID only supported in PercentVbus mode")

    def set(self, outputValue, syncGroup=0):
        """
        Sets the appropriate output on the talon, depending on the mode.

        In PercentVbus, the output is between -1.0 and 1.0, with 0.0 as stopped.

        In Follower mode, the output is the integer device ID of the talon to
            duplicate.

        In Voltage mode, outputValue is in volts.

        In Current mode, outputValue is in amperes.

        In Speed mode, outputValue is in position change / 10ms.

        In Position mode, outputValue is in encoder ticks or an analog value,
            depending on the sensor.

        :param outputValue: The setpoint value, as described above.
        """
        if not self.controlEnabled:
            return
        self.setPoint = outputValue
        if self.controlMode == self.ControlMode.PercentVbus:
            if outputValue > 1:
                outputValue = 1
            elif outputValue < -1:
                outputValue = -1
            hal.TalonSRX_SetDemand(self.handle, int(1023*outputValue))
        elif self.controlMode == self.ControlMode.Follower:
            hal.TalonSRX_SetDemand(self.handle, int(outputValue))
        elif self.controlMode == self.ControlMode.Voltage:
            # Voltage is an 8.8 fixed point number.
            volts = int(outputValue * 256)
            hal.TalonSRX_SetDemand(self.handle, volts)
        elif self.controlMode == self.ControlMode.Speed:
            hal.TalonSRX_SetDemand(self.handle, int(outputValue))
        elif self.controlMode == self.ControlMode.Position:
            hal.TalonSRX_SetDemand(self.handle, int(outputValue))
        hal.TalonSRX_SetModeSelect(self.handle, self.controlMode)

    def reverseSensor(self, flip):
        """
        Flips the sign (multiplies by negative one) the sensor values going
        into the talon.

        This only affects position and velocity closed loop control. Allows for
          situations where you may have a sensor flipped and going in the wrong
          direction.

        :param flip: True if sensor input should be flipped; False if not.
        """
        hal.TalonSRX_SetRevFeedbackSensor(self.handle, 1 if flip else 0)

    def reverseOutput(self, flip):
        """
        Flips the sign (multiplies by negative one) the throttle values going
        into the motor on the talon in closed loop modes.

        :param flip: True if motor output should be flipped; False if not.
        """
        hal.TalonSRX_SetRevMotDuringCloseLoopEn(self.handle, 1 if flip else 0)

    def get(self):
        """
        Gets the current status of the Talon (usually a sensor value).

        In Current mode: returns output current.

        In Speed mode: returns current speed.

        In Position omde: returns current sensor position.

        In Throttle and Follower modes: returns current applied throttle.

        :returns: The current sensor value of the Talon.
        """
        if self.controlMode == self.ControlMode.Voltage:
            return self.getOutputVoltage()
        elif self.controlMode == self.ControlMode.Current:
            return self.getOutputCurrent()
        elif self.controlMode == self.ControlMode.Speed:
            return float(hal.TalonSRX_GetSensorVelocity(self.handle))
        elif self.controlMode == self.ControlMode.Position:
            return float(hal.TalonSRX_GetSensorPosition(self.handle))
        else: # PercentVbus
            return hal.TalonSRX_GetAppliedThrottle(self.handle) / 1023.0

    def getEncPosition(self):
        """
        Get the current encoder position, regardless of whether it is the
        current feedback device.

        :returns: The current position of the encoder.
        """
        return hal.TalonSRX_GetEncPosition(self.handle)

    def getEncVelocity(self):
        """
        Get the current encoder velocity, regardless of whether it is the
        current feedback device.

        :returns: The current speed of the encoder.
        """
        return hal.TalonSRX_GetEncVel(self.handle)

    def getNumberOfQuadIdxRises(self):
        """
        Get the number of of rising edges seen on the index pin.

        :returns: number of rising edges on idx pin.
        """
        return hal.TalonSRX_GetEncIndexRiseEvents(self.handle)

    def getPinStateQuadA(self):
        """
        :returns: IO level of QUADA pin.
        """
        return hal.TalonSRX_GetQuadApin(self.handle)

    def getPinStateQuadB(self):
        """
        :returns: IO level of QUADB pin.
        """
        return hal.TalonSRX_GetQuadBpin(self.handle)

    def getPinStateQuadIdx(self):
        """
        :returns: IO level of QUAD Index pin.
        """
        return hal.TalonSRX_GetQuadIdxpin(self.handle)

    def getAnalogInPosition(self):
        """
        Get the current analog in position, regardless of whether it is the
        current feedback device.

        :returns: The 24bit analog position.  The bottom ten bits is the ADC
            (0 - 1023) on the analog pin of the Talon. The upper 14 bits
            tracks the overflows and underflows (continuous sensor).
        """
        return hal.TalonSRX_GetAnalogInWithOv(self.handle)

    def getAnalogInRaw(self):
        """
        Get the current analog in position, regardless of whether it is the
        current feedback device.
        :returns: The ADC (0 - 1023) on analog pin of the Talon.
        """
        return self.getAnalogInPosition() & 0x3FF

    def getAnalogInVelocity(self):
        """
        Get the current encoder velocity, regardless of whether it is the
        current feedback device.

        :returns: The current speed of the analog in device.
        """
        return hal.TalonSRX_GetAnalogInVel(self.handle)

    def getClosedLoopError(self):
        """
        Get the current difference between the setpoint and the sensor value.

        :returns: The error, in whatever units are appropriate.
        """
        return hal.TalonSRX_GetCloseLoopErr(self.handle)

    def isFwdLimitSwitchClosed(self):
        """Returns True if limit switch is closed. False if open."""
        return hal.TalonSRX_GetLimitSwitchClosedFor(self.handle) == 0

    def isRevLimitSwitchClosed(self):
        """Returns True if limit switch is closed. False if open."""
        return hal.TalonSRX_GetLimitSwitchClosedRev(self.handle) == 0

    def getBrakeEnableDuringNeutral(self):
        """Returns True if break is enabled during neutral. False if coast."""
        return hal.TalonSRX_GetBrakeIsEnabled(self.handle) != 0

    def getTemp(self):
        """Returns temperature of Talon, in degrees Celsius."""
        return hal.TalonSRX_GetTemp(self.handle)

    def getSensorPosition(self):
        return hal.TalonSRX_GetSensorPosition(self.handle)

    def getSensorVelocity(self):
        return hal.TalonSRX_GetSensorVelocity(self.handle)

    def getOutputCurrent(self):
        """Returns the current going through the Talon, in Amperes."""
        return hal.TalonSRX_GetCurrent(self.handle)

    def getOutputVoltage(self):
        """
        :returns: The voltage being output by the Talon, in Volts.
        """
        return self.getBusVoltage() * hal.TalonSRX_GetAppliedThrottle(self.handle) / 1023.0

    def getBusVoltage(self):
        """
        :returns: The voltage at the battery terminals of the Talon, in Volts.
        """
        return hal.TalonSRX_GetBatteryV(self.handle)

    def getPosition(self):
        return hal.TalonSRX_GetSensorPosition(self.handle)

    def setPosition(self, pos):
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eSensorPosition,
                              int(pos))

    def getSpeed(self):
        return hal.TalonSRX_GetSensorVelocity(self.handle)

    def getControlMode(self):
        return self.controlMode

    def _applyControlMode(self, controlMode):
        """
        Fixup the self.controlMode so set() serializes the correct demand value.
        Also fills the modeSelecet in the control frame to disabled.
        :param controlMode: Control mode to ultimately enter once user calls
            set().
        """
        self.controlMode = controlMode
        if controlMode == self.ControlMode.Disabled:
            self.controlEnabled = False
        # Disable until set() is called.
        hal.TalonSRX_SetModeSelect(self.handle, self.ControlMode.Disabled)

    def changeControlMode(self, controlMode):
        # if we already are in this mode, don't perform disable workaround
        if self.controlMode != controlMode:
            self._applyControlMode(controlMode)

    def setFeedbackDevice(self, device):
        hal.TalonSRX_SetFeedbackDeviceSelect(self.handle, device)

    def setStatusFrameRateMs(self, stateFrame, periodMs):
        """Change the periodMs of a TALON's status frame.  See StatusFrameRate
        enum for what's available."""
        hal.TalonSRX_SetStatusFrameRate(self.handle, stateFrame, periodMs)

    def enableControl(self):
        self.changeControlMode(self.controlMode)
        self.controlEnabled = True

    def disableControl(self):
        hal.TalonSRX_SetModeSelect(self.handle, self.ControlMode.Disabled)
        self.controlEnabled = False

    def isControlEnabled(self):
        return self.controlEnabled

    def _getParam(self, paramEnum):
        # Update the information that we have.
        hal.TalonSRX_RequestParam(self.handle, paramEnum)

        # Briefly wait for new values from the Talon.
        Timer.delay(self.kDelayForSolicitedSignals)

        return hal.TalonSRX_GetParamResponse(self.handle, paramEnum)

    def _getParamInt(self, paramEnum):
        # Update the information that we have.
        hal.TalonSRX_RequestParam(self.handle, paramEnum)

        # Briefly wait for new values from the Talon.
        Timer.delay(self.kDelayForSolicitedSignals)

        return hal.TalonSRX_GetParamResponseInt32(self.handle, paramEnum)

    def getP(self):
        """
        Get the current proportional constant.

        :returns: double proportional constant for current profile.
        """
        if self.profile == 0:
            return self._getParam(hal.TalonSRXParam.eProfileParamSlot0_P)
        else:
            return self._getParam(hal.TalonSRXParam.eProfileParamSlot1_P)

    def getI(self):
        if self.profile == 0:
            return self._getParam(hal.TalonSRXParam.eProfileParamSlot0_I)
        else:
            return self._getParam(hal.TalonSRXParam.eProfileParamSlot1_I)

    def getD(self):
        if self.profile == 0:
            return self._getParam(hal.TalonSRXParam.eProfileParamSlot0_D)
        else:
            return self._getParam(hal.TalonSRXParam.eProfileParamSlot1_D)

    def getF(self):
        if self.profile == 0:
            return self._getParam(hal.TalonSRXParam.eProfileParamSlot0_F)
        else:
            return self._getParam(hal.TalonSRXParam.eProfileParamSlot1_F)

    def getIZone(self):
        if self.profile == 0:
            return self._getParamInt(hal.TalonSRXParam.eProfileParamSlot0_IZone)
        else:
            return self._getParamInt(hal.TalonSRXParam.eProfileParamSlot1_IZone)

    def getCloseLoopRampRate(self):
        """
        Get the closed loop ramp rate for the current profile.

        Limits the rate at which the throttle will change.
        Only affects position and speed closed loop modes.

        :returns: rampRate Maximum change in voltage, in volts / sec.

        :see: #setProfile For selecting a certain profile.
        """
        if self.profile == 0:
            throttlePerMs = self._getParamInt(hal.TalonSRXParam.eProfileParamSlot0_CloseLoopRampRate)
        else:
            throttlePerMs = self._getParamInt(hal.TalonSRXParam.eProfileParamSlot1_CloseLoopRampRate)

        return throttlePerMs / 1023.0 * 12.0 * 1000.0

    def getFirmwareVersion(self):
        """
        :returns: The version of the firmware running on the Talon
        """
        return self._getParamInt(hal.TalonSRXParam.eFirmVers)

    def getIaccum(self):
        return self._getParamInt(hal.TalonSRXParam.ePidIaccum)

    def clearIaccum(self):
        """
        Clear the accumulator for I gain.
        """
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.ePidIaccum, 0)

    def setP(self, p):
        """
        Set the proportional value of the currently selected profile.

        :param p: Proportional constant for the currently selected PID profile.
        :see: #setProfile For selecting a certain profile.
        """
        if self.profile == 0:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot0_P, p)
        else:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot1_P, p)

    def setI(self, i):
        """
        Set the integration constant of the currently selected profile.

        :param i: Integration constant for the currently selected PID profile.
        :see: #setProfile For selecting a certain profile.
        """
        if self.profile == 0:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot0_I, i)
        else:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot1_I, i)

    def setD(self, d):
        """
        Set the derivative constant of the currently selected profile.

        :param d: Derivative constant for the currently selected PID profile.
        :see: #setProfile For selecting a certain profile.
        """
        if self.profile == 0:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot0_D, d)
        else:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot1_D, d)

    def setF(self, f):
        """
        Set the feedforward value of the currently selected profile.

        :param f: Feedforward constant for the currently selected PID profile.
        :see: #setProfile For selecting a certain profile.
        """
        if self.profile == 0:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot0_F, f)
        else:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot1_F, f)

    def setIZone(self, izone):
        """
        Set the integration zone of the current Closed Loop profile.

        Whenever the error is larger than the izone value, the accumulated
        integration error is cleared so that high errors aren't racked up when
        at high errors.

        An izone value of 0 means no difference from a standard PIDF loop.

        :param izone: Width of the integration zone.
        :see: #setProfile For selecting a certain profile.
        """
        if self.profile == 0:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot0_IZone, izone)
        else:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot1_IZone, izone)

    def setCloseLoopRampRate(self, rampRate):
        """
        Set the closed loop ramp rate for the current profile.

        Limits the rate at which the throttle will change.
        Only affects position and speed closed loop modes.

        :param rampRate: Maximum change in voltage, in volts / sec.
        :see: #setProfile For selecting a certain profile.
        """
        # CanTalonSRX takes units of Throttle (0 - 1023) / 1ms.
        rate = int(rampRate * 1023.0 / 12.0 / 1000.0)
        if self.profile == 0:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot0_CloseLoopRampRate, rate)
        else:
            hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSlot1_CloseLoopRampRate, rate)

    def setVoltageRampRate(self, rampRate):
        """
        Set the voltage ramp rate for the current profile.

        Limits the rate at which the throttle will change.
        Affects all modes.

        :param rampRate: Maximum change in voltage, in volts / sec.
        """
        # CanTalonSRX takes units of Throttle (0 - 1023) / 10ms.
        rate = int(rampRate * 1023.0 / 12.0 /100.0)
        hal.TalonSRX_SetRampThrottle(self.handle, rate)

    def setPID(self, p, i, d, f=0, izone=0, closeLoopRampRate=0,
               profile=None):
        """
        Sets control values for closed loop control.

        :param p: Proportional constant.
        :param i: Integration constant.
        :param d: Differential constant.
        :param f: Feedforward constant.
        :param izone: Integration zone -- prevents accumulation of integration
            error with large errors. Setting this to zero will ignore any
            izone stuff.
        :param closeLoopRampRate: Closed loop ramp rate. Maximum change in
            voltage, in volts / sec.
        :param profile: which profile to set the pid constants for. You can
            have two profiles, with values of 0 or 1, allowing you to keep a
            second set of values on hand in the talon. In order to switch
            profiles without recalling setPID, you must call setProfile().
        """
        if profile is not None:
            self.setProfile(profile)
        self.setP(p)
        self.setI(i)
        self.setD(d)
        self.setF(f)
        self.setIZone(izone)
        self.setCloseLoopRampRate(closeLoopRampRate)

    def getSetpoint(self):
        """
        :returns: The latest value set using set().
        """
        return self.setPoint

    def setProfile(self, profile):
        """
        Select which closed loop profile to use, and uses whatever PIDF gains
        and the such that are already there.
        """
        if profile not in (0, 1):
            raise ValueError("Talon PID profile must be 0 or 1.")
        self.profile = profile
        hal.TalonSRX_SetProfileSlotSelect(self.handle, self.profile)

    def stopMotor(self):
        """
        Common interface for stopping a motor.
        """
        self.disableControl()

    def disable(self):
        self.disableControl()

    def getDeviceID(self):
        return self.deviceNumber

    # TODO: Documentation for all these accessors/setters for misc. stuff.

    def setSensorPosition(self, pos):
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eSensorPosition, pos)

    def setForwardSoftLimit(self, forwardLimit):
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSoftLimitForThreshold, forwardLimit)

    def enableForwardSoftLimit(self, enable):
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSoftLimitForEnable, 1 if enable else 0)

    def setReverseSoftLimit(self, reverseLimit):
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSoftLimitRevThreshold, reverseLimit)

    def enableReverseSoftLimit(self, enable):
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eProfileParamSoftLimitRevEnable, 1 if enable else 0)

    def clearStickyFaults(self):
        hal.TalonSRX_ClearStickyFaults(self.handle)

    def enableLimitSwitch(self, forward, reverse):
        mask = 4 + (2 if forward else 0) + (1 if reverse else 0)
        hal.TalonSRX_SetOverrideLimitSwitchEn(self.handle, mask)

    def configFwdLimitSwitchNormallyOpen(self, normallyOpen):
        """
        Configure the fwd limit switch to be normally open or normally closed.
        Talon will disable momentarilly if the Talon's current setting
        is dissimilar to the caller's requested setting.

        Since Talon saves setting to flash this should only affect
        a given Talon initially during robot install.

        :param normallyOpen: True for normally open. False for normally closed.
        """
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eOnBoot_LimitSwitch_Forward_NormallyClosed, 0 if normallyOpen else 1)

    def configRevLimitSwitchNormallyOpen(self, normallyOpen):
        """
        * Configure the rev limit switch to be normally open or normally closed.
        * Talon will disable momentarilly if the Talon's current setting
        * is dissimilar to the caller's requested setting.
        *
        * Since Talon saves setting to flash this should only affect
        * a given Talon initially during robot install.
        *
        * @param normallyOpen true for normally open.  false for normally closed.
        """
        hal.TalonSRX_SetParam(self.handle, hal.TalonSRXParam.eOnBoot_LimitSwitch_Reverse_NormallyClosed, 0 if normallyOpen else 1)

    def enableBrakeMode(self, brake):
        hal.TalonSRX_SetOverrideBrakeType(self.handle, 2 if brake else 1)

    def getFaultOverTemp(self):
        return hal.TalonSRX_GetFault_OverTemp(self.handle)

    def getFaultUnderVoltage(self):
        return hal.TalonSRX_GetFault_UnderVoltage(self.handle)

    def getFaultForLim(self):
        return hal.TalonSRX_GetFault_ForLim(self.handle)

    def getFaultRevLim(self):
        return hal.TalonSRX_GetFault_RevLim(self.handle)

    def getFaultHardwareFailure(self):
        return hal.TalonSRX_GetFault_HardwareFailure(self.handle)

    def getFaultForSoftLim(self):
        return hal.TalonSRX_GetFault_ForSoftLim(self.handle)

    def getFaultRevSoftLim(self):
        return hal.TalonSRX_GetFault_RevSoftLim(self.handle)

    def getStickyFaultOverTemp(self):
        return hal.TalonSRX_GetStckyFault_OverTemp(self.handle)

    def getStickyFaultUnderVoltage(self):
        return hal.TalonSRX_GetStckyFault_UnderVoltage(self.handle)

    def getStickyFaultForLim(self):
        return hal.TalonSRX_GetStckyFault_ForLim(self.handle)

    def getStickyFaultRevLim(self):
        return hal.TalonSRX_GetStckyFault_RevLim(self.handle)

    def getStickyFaultForSoftLim(self):
        return hal.TalonSRX_GetStckyFault_ForSoftLim(self.handle)

    def getStickyFaultRevSoftLim(self):
        return hal.TalonSRX_GetStckyFault_RevSoftLim(self.handle)

    def getDescription(self):
        return "CANTalon ID %d" % self.deviceNumber
    
    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Speed Controller"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.get())

    def valueChanged(self, itable, key, value, bln):
        self.set(float(value))

    def startLiveWindowMode(self):
        self.set(0) # Stop for safety
        super().startLiveWindowMode()

    def stopLiveWindowMode(self):
        super().stopLiveWindowMode()
        self.set(0) # Stop for safety
    



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

__all__ = ["I2C"]

def _freeI2C(port):
    hal.i2CClose(port)

class I2C:
    """I2C bus interface class.

    This class is intended to be used by sensor (and other I2C device) drivers.
    It probably should not be used directly.
    """
    class Port:
        kOnboard = 0
        kMXP = 1

    def __init__(self, port, deviceAddress):
        """Constructor.

        :param port: The I2C port the device is connected to.
        :param deviceAddress: The address of the device on the I2C bus.
        """
        self.port = port
        self.deviceAddress = deviceAddress

        hal.i2CInitialize(self.port)
        self._i2c_finalizer = weakref.finalize(self, _freeI2C, self.port)

        hal.HALReport(hal.HALUsageReporting.kResourceType_I2C, deviceAddress)

    def transaction(self, dataToSend, receiveSize):
        """Generic transaction.

        This is a lower-level interface to the I2C hardware giving you more
        control over each transaction.

        :param dataToSend:
            Data to send as part of the transaction.
        :param receiveSize:
            Number of bytes to read from the device. [0..7]
        :returns: Data received from the device.
        """
        return hal.i2CTransaction(self.port, self.deviceAddress,
                                  dataToSend, receiveSize)

    def addressOnly(self):
        """Attempt to address a device on the I2C bus.

        This allows you to figure out if there is a device on the I2C bus that
        responds to the address specified in the constructor.

        :returns: Transfer Aborted... False for success, True for aborted.
        """
        try:
            self.transaction([], 0)
        except IOError:
            return True
        return False

    def write(self, registerAddress, data):
        """Execute a write transaction with the device.

        Write a single byte to a register on a device and wait until the
        transaction is complete.

        :param registerAddress:
            The address of the register on the device to be written.
        :param data: The byte to write to the register on the device.
        :returns: Transfer Aborted... False for success, True for aborted.
        """
        try:
            hal.i2CWrite(self.port, self.deviceAddress, [registerAddress, data])
        except IOError:
            return True
        return False

    def writeBulk(self, data):
        """Execute a write transaction with the device.

        Write multiple bytes to a register on a device and wait until the
        transaction is complete.

        :param data: The data to write to the device.
        :returns: Transfer Aborted... False for success, True for aborted.
        """
        try:
            hal.i2CWrite(self.port, self.deviceAddress, data)
        except IOError:
            return True
        return False

    def read(self, registerAddress, count):
        """Execute a read transaction with the device.

        Read 1 to 7 bytes from a device. Most I2C devices will auto-increment
        the register pointer internally allowing you to read up to 7
        consecutive registers on a device in a single transaction.

        :param registerAddress: The register to read first in the transaction.
        :param count: The number of bytes to read in the transaction. [1..7]
        :returns: The data read from the device.
        """
        if count < 1 or count > 7:
            raise ValueError("count must be between 1 and 7")
        return self.transaction([registerAddress], count)

    def readOnly(self, count):
        """Execute a read only transaction with the device.

        Read 1 to 7 bytes from a device. This method does not write any data
        to prompt the device.

        :param count: The number of bytes to read in the transaction. [1..7]
        :returns: The data read from the device.
        """
        if count < 1 or count > 7:
            raise ValueError("count must be between 1 and 7")
        hal.i2CRead(self.port, self.deviceAddress, count)

    def broadcast(self, registerAddress, data):
        """Send a broadcast write to all devices on the I2C bus.

        .. warning:: This is not currently implemented!

        :param registerAddress:
            The register to write on all devices on the bus.
        :param data: The value to write to the devices.
        """
        raise NotImplementedError

    def verifySensor(self, registerAddress, expected):
        """Verify that a device's registers contain expected values.

        Most devices will have a set of registers that contain a known value
        that can be used to identify them. This allows an I2C device driver
        to easily verify that the device contains the expected value.

        The device must support and be configured to use register
        auto-increment.

        :param registerAddress:
            The base register to start reading from the device.
        :param expected: The values expected from the device.
        :returns: True if the sensor was verified to be connected
        """
        # TODO: Make use of all 7 read bytes
        count = len(expected)
        for i in range(0, count, 4):
            toRead = 4
            if (count - i) < 4:
                toRead = count - i
            # Read the chunk of data. Return False if the sensor does not
            # respond.
            try:
                deviceData = self.read(registerAddress, toRead)
            except IOError:
                return False

            for j in range(toRead):
                if deviceData[j] != expected[i + j]:
                    return False

            registerAddress += count

        return True



"""
    This package contains objects that can be used to determine the
    requirements of various interfaces used in WPILib. 
    
    Generally, the python version of WPILib does not require that you inherit
    from any of these interfaces, but instead will allow you to use custom 
    objects as long as they have the same methods.
"""

from .accelerometer import *
from .controller import *
from .counterbase import *
from .generichid import *
from .namedsendable import *
from .pidoutput import *
from .pidsource import *
from .potentiometer import *
from .speedcontroller import *



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .livewindow import LiveWindow
from .safepwm import SafePWM

__all__ = ["VictorSP"]

class VictorSP(SafePWM):
    """
        VEX Robotics Victor SP Speed Controller via PWM
        
        .. not_implemented: initVictorSP
    """

    def __init__(self, channel):
        """Constructor.

        :param channel: The PWM channel that the VictorSP is attached to. 0-9 are on-board, 10-19 are on the MXP port.
        :type  channel: int

        .. note ::

            The Talon uses the following bounds for PWM values. These values
            should work reasonably well for most controllers, but if users
            experience issues such as asymmetric behavior around the deadband
            or inability to saturate the controller in either direction,
            calibration is recommended.  The calibration procedure can be
            found in the VictorSP User Manual.

            - 2.004ms = full "forward"
            - 1.520ms = the "high end" of the deadband range
            - 1.500ms = center of the deadband range (off)
            - 1.480ms = the "low end" of the deadband range
            - 0.997ms = full "reverse"
        """
        super().__init__(channel)
        self.setBounds(2.004, 1.52, 1.50, 1.48, .997)
        self.setPeriodMultiplier(self.PeriodMultiplier.k1X)
        self.setRaw(self.centerPwm)
        self.setZeroLatch()

        LiveWindow.addActuatorChannel("VictorSP", self.getChannel(), self)
        hal.HALReport(hal.HALUsageReporting.kResourceType_VictorSP,
                      self.getChannel())

    def set(self, speed, syncGroup=0):
        """Set the PWM value.

        The PWM value is set using a range of -1.0 to 1.0, appropriately
        scaling the value for the FPGA.

        :param speed: The speed to set.  Value should be between -1.0 and 1.0.
        :type  speed: float
        :param syncGroup: The update group to add this set() to, pending
            updateSyncGroup().  If 0, update immediately.
        """
        self.setSpeed(speed)
        self.feed()

    def get(self):
        """Get the recently set value of the PWM.

        :returns: The most recently set value for the PWM between -1.0 and 1.0.
        :rtype: float
        """
        return self.getSpeed()

    def pidWrite(self, output):
        """Write out the PID value as seen in the PIDOutput base object.

        :param output: Write out the PWM value as was found in the
            :class:`PIDController`.
        :type  output: float
        """
        self.set(output)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .trigger import Trigger

__all__ = ["Button"]

class Button(Trigger):
    """This class provides an easy way to link commands to OI inputs.

    It is very easy to link a button to a command.  For instance, you could
    link the trigger button of a joystick to a "score" command.

    This class represents a subclass of :class:`.Trigger` that is specifically aimed at
    buttons on an operator interface as a common use case of the more
    generalized Trigger objects. This is a simple wrapper around Trigger with
    the method names renamed to fit the Button object use.
    """

    def whenPressed(self, command):
        """Starts the given command whenever the button is newly pressed.

        :param command: the command to start
        """
        self.whenActive(command)

    def whileHeld(self, command):
        """Constantly starts the given command while the button is held.

        :meth:`.Command.start` will be called repeatedly while the button is
        held, and will be canceled when the button is released.

        :param command: the command to start
        """
        self.whileActive(command)

    def whenReleased(self, command):
        """Starts the command when the button is released.

        :param command: the command to start
        """
        self.whenInactive(command)

    def toggleWhenPressed(self, command):
        """Toggles the command whenever the button is pressed (on then off
        then on).

        :param command:
        """
        self.toggleWhenActive(command)

    def cancelWhenPressed(self, command):
        """Cancel the command when the button is pressed.

        :param command:
        """
        self.cancelWhenActive(command)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .livewindow import LiveWindow
from .safepwm import SafePWM

__all__ = ["Jaguar"]

class Jaguar(SafePWM):
    """
        Texas Instruments / Vex Robotics Jaguar Speed Controller as a PWM device.
    
        .. seealso::
        
           :class:`.CANJaguar` for CAN control of a Jaguar
           
        .. not_implemented: initJaguar
    """

    def __init__(self, channel):
        """Constructor.

        :param channel: The PWM channel that the Jaguar is attached to. 0-9 are on-board, 10-19 are on the MXP port
        """
        super().__init__(channel)
        # Input profile defined by Luminary Micro.
        #
        # Full reverse ranges from 0.671325ms to 0.6972211ms
        # Proportional reverse ranges from 0.6972211ms to 1.4482078ms
        # Neutral ranges from 1.4482078ms to 1.5517922ms
        # Proportional forward ranges from 1.5517922ms to 2.3027789ms
        # Full forward ranges from 2.3027789ms to 2.328675ms
        self.setBounds(2.31, 1.55, 1.507, 1.454, 0.697)
        self.setPeriodMultiplier(self.PeriodMultiplier.k1X)
        self.setRaw(self.centerPwm)
        self.setZeroLatch()

        hal.HALReport(hal.HALUsageReporting.kResourceType_Jaguar,
                      self.getChannel())
        LiveWindow.addActuatorChannel("Jaguar", self.getChannel(), self)

    def set(self, speed, syncGroup=0):
        """Set the PWM value.

        The PWM value is set using a range of -1.0 to 1.0, appropriately
        scaling the value for the FPGA.

        :param speed: The speed to set.  Value should be between -1.0 and 1.0.
        :type  speed: float
        :param syncGroup: The update group to add this set() to, pending
            updateSyncGroup().  If 0, update immediately.
        """
        self.setSpeed(speed)
        self.feed()

    def get(self):
        """Get the recently set value of the PWM.

        :returns: The most recently set value for the PWM between -1.0 and 1.0.
        :rtype: float
        """
        return self.getSpeed()

    def pidWrite(self, output):
        """Write out the PID value as seen in the PIDOutput base object.

        :param output: Write out the PWM value as was found in the
            :class:`PIDController`.
        :type  output: float
        """
        self.set(output)
